"""Aerotech"""
import time
from concert.quantities import q
from concert.networking.aerotech import Connection
from concert.devices.motors.base import ContinuousRotationMotor


class Aerorot(ContinuousRotationMotor):

    """Aerorot ContinuousRotationMotor class implementation."""
    AXIS = "X"

    # status constants (bits of the AXISSTATUS output (see HLe docs))
    AXISSTATUS_ENABLED = 0
    AXISSTATUS_HOMED = 1
    AXISSTATUS_IN_POSITION = 2
    AXISSTATUS_MOVE_ACTIVE = 3
    AXISSTATUS_ACCEL_PHASE = 4
    AXISSTATUS_DECEL_PHASE = 5
    AXISSTATUS_POSITION_CAPTURE = 6
    AXISSTATUS_HOMING = 14

    SLEEP_TIME = 0.01

    def __init__(self, host, port=8001, enable=True):
        super(Aerorot, self).__init__()

        self._connection = Connection(host, port)
        if enable:
            self.enable()

    def enable(self):
        """Enable the motor."""
        self._connection.execute("ENABLE %s" % (Aerorot.AXIS))

    def disable(self):
        """Disable the motor."""
        self._connection.execute("DISABLE %s" % (Aerorot.AXIS))

    def _query_state(self):
        return int(self._connection.execute("AXISSTATUS(%s)" % (Aerorot.AXIS)))

    def _get_position(self):
        return float(self._connection.execute("PFBK(%s)" % (Aerorot.AXIS))) * q.deg

    def _set_position(self, position):
        self._connection.execute("MOVEABS %s %f" % (Aerorot.AXIS, position.magnitude))

        while not self._query_state() >> Aerorot.AXISSTATUS_IN_POSITION & 1:
            time.sleep(Aerorot.SLEEP_TIME)

    def _get_velocity(self):
        return float(self._connection.execute("VFBK(%s)" % (Aerorot.AXIS))) * q.deg / q.s

    def _set_velocity(self, velocity):
        self._connection.execute("FREERUN %s %f" % (Aerorot.AXIS, velocity.magnitude))

        while self._query_state() >> Aerorot.AXISSTATUS_ACCEL_PHASE & 1:
            time.sleep(Aerorot.SLEEP_TIME)

    def check_state(self):
        res = self._query_state()

        # Simplified behavior because of unstable motor states, i.e.
        # error after long-homing etx. TODO:investigate
        state = 'standby'

        if res >> Aerorot.AXISSTATUS_MOVE_ACTIVE & 1:
            state = 'moving'

        return state

    def _stop(self):
        if self.check_state() == 'moving':
            self._connection.execute("ABORT %s" % (Aerorot.AXIS))

        while self.check_state() == 'moving':
            time.sleep(Aerorot.SLEEP_TIME)

    def _home(self):
        self._connection.execute("HOME %s" % (Aerorot.AXIS))

        while self.check_state() == 'moving':
            time.sleep(Aerorot.SLEEP_TIME)



"""
A :class:`Camera` can be set via the device-specific properties that can be set
and read with :meth:`.Parameter.set` and :meth:`.Parameter.get`.  Moreover, a
camera provides means to

* :meth:`~Camera.start_recording` frames,
* :meth:`~Camera.stop_recording` the acquisition,
* :meth:`~Camera.trigger` a frame capture and
* :meth:`~Camera.grab` to get the last frame.

Camera triggering is specified by the :attr:`~Camera.trigger_mode` parameter, which
can be one of

* :attr:`camera.trigger_modes.AUTO` means the camera triggers itself
  automatically, the frames start being recorded right after the
  :meth:`~camera.start_recording` call and stop being recorded by
  :meth:`~camera.stop_recording`

* :attr:`Camera.trigger_modes.SOFTWARE` means the camera needs to be triggered
  by the user by :meth:`~Camera.trigger()`. This way you have complete programatic
  control over when is the camera triggered, example usage::

    camera.trigger_mode = camera.trigger_modes.SOFTWARE
    camera.start_recording()
    camera.trigger()
    long_operation()
    # Here we get the frame from before the long operation
    camera.grab()

* :attr:`Camera.trigger_modes.EXTERNAL` is a mode when the camera is triggered
  by an external low-level signal (such as TTL). This mode provides very precise
  triggering in terms of time synchronization with other devices


To setup and use a camera in a typical environment, you would do::

    import numpy as np
    from concert.devices.cameras.uca import UcaCamera

    camera = UcaCamera('pco')
    camera.trigger_mode = camera.trigger_modes.SOFTWARE
    camera.exposure_time = 0.2 * q.s
    camera.start_recording()
    camera.trigger()
    data = camera.grab()
    camera.stop_recording()

    print("mean=%f, stddev=%f" % (np.mean(data), np.std(data))
"""
import contextlib
from concert.base import AccessorNotImplementedError, Parameter, Quantity, State, transition
from concert.async import async
from concert.quantities import q
from concert.helpers import Bunch
from concert.devices.base import Device


class CameraError(Exception):

    """Camera specific errors."""
    pass


class Camera(Device):

    """Base class for remotely controllable cameras.

    .. py:attribute:: frame-rate

        Frame rate of acquisition in q.count per time unit.
    """

    trigger_modes = Bunch(['AUTO', 'SOFTWARE', 'EXTERNAL'])
    state = State(default='standby')
    frame_rate = Quantity(1.0 / q.second, help="Frame frequency")
    trigger_mode = Parameter(help="Trigger mode")

    def __init__(self):
        super(Camera, self).__init__()

    @transition(source='standby', target='recording')
    def start_recording(self):
        """Start recording frames."""
        self._record_real()

    @transition(source='recording', target='standby')
    def stop_recording(self):
        """Stop recording frames."""
        self._stop_real()

    @contextlib.contextmanager
    def recording(self):
        """
        A context manager for starting and stopping the camera.

        In general it is used with the ``with`` keyword like this::

            with camera.recording():
                frame = camera.grab()
        """
        self.start_recording()
        try:
            yield
        finally:
            self.stop_recording()

    def trigger(self):
        """Trigger a frame if possible."""
        self._trigger_real()

    def grab(self):
        """Return a NumPy array with data of the current frame."""
        return self._grab_real()

    @async
    def stream(self, consumer):
        """Grab frames continuously and send them to *consumer*, which
        is a coroutine.
        """
        self.trigger_mode = self.trigger_modes.AUTO
        self.start_recording()

        while self.state == 'recording':
            consumer.send(self.grab())

    def _get_trigger_mode(self):
        raise AccessorNotImplementedError

    def _set_trigger_mode(self, mode):
        raise AccessorNotImplementedError

    def _record_real(self):
        raise AccessorNotImplementedError

    def _stop_real(self):
        raise AccessorNotImplementedError

    def _trigger_real(self):
        raise AccessorNotImplementedError

    def _grab_real(self):
        raise AccessorNotImplementedError


class BufferedMixin(Device):

    """A camera that stores the frames in an internal buffer"""

    state = State(default='standby')

    @transition(source='standby', target='standby', immediate='readout')
    def readout_buffer(self, *args, **kwargs):
        return self._readout_real(*args, **kwargs)

    def _readout_real(self, *args, **kwargs):
        raise AccessorNotImplementedError



import os
from concert.storage import create_directory, write_tiff
from .base import coroutine


class Result(object):
    """
    The object is callable and when called it becomes a coroutine which accepts
    items and stores them in a variable which allows the user to obtain the
    last stored item at any time point.
    """
    def __init__(self):
        self.result = None

    @coroutine
    def __call__(self):
        """
        __call__(self)

        When the object is called we store every item in an attribute which
        can be recovered later.
        """
        while True:
            self.result = yield


@coroutine
def null():
    """
    null()

    A black-hole.
    """
    while True:
        yield


@coroutine
def write_images(writer=write_tiff, prefix="image_{:>05}"):
    """
    write_images(writer, prefix="image_{:>05}")

    Write images on disk with specified *writer* and file name *prefix*.
    *writer* is a callable with the following nomenclature::

        writer(file_name_prefix, data)

    The file extension needs to be applied by a particular writer.
    """
    i = 0
    dir_name = os.path.dirname(prefix)

    if dir_name and not os.path.exists(dir_name):
        create_directory(dir_name)

    while True:
        data = yield
        writer(prefix.format(i), data)
        i += 1



"""Pumps."""

from concert.base import Quantity, State, transition, AccessorNotImplementedError
from concert.quantities import q
from concert.async import async
from concert.devices.base import Device


class Pump(Device):

    """A pumping device."""

    state = State(default='standby')
    flow_rate = Quantity(q.l / q.s, help="Flow rate")

    def __init__(self):
        super(Pump, self).__init__()

    @async
    @transition(source='standby', target='pumping')
    def start(self):
        """
        start()

        Start pumping.
        """
        self._start()

    @async
    @transition(source='pumping', target='standby')
    def stop(self):
        """
        stop()

        Stop pumping.
        """
        self._stop()

    def _get_flow_rate(self):
        raise AccessorNotImplementedError

    def _set_flow_rate(self, flow_rate):
        raise AccessorNotImplementedError

    def _start(self):
        raise NotImplementedError

    def _stop(self):
        raise NotImplementedError



from concert.base import Parameter, ParameterError
from concert.devices.base import Device
from concert.tests import TestCase, suppressed_logging
from concert.devices.scales.dummy import Scales, TarableScales
from concert.devices.pumps.dummy import Pump
from concert.devices.io.dummy import IO
from concert.devices.monochromators.dummy import Monochromator
from concert.devices.shutters.dummy import Shutter
from concert.devices.storagerings.dummy import StorageRing
from concert.devices.photodiodes.dummy import PhotoDiode
from concert.devices.lightsources.dummy import LightSource


@suppressed_logging
def test_dummies():
    Scales()
    TarableScales()
    Pump()
    IO()
    Monochromator()
    Shutter()
    StorageRing()
    PhotoDiode()
    LightSource()


class MockDevice(Device):

    readonly = Parameter()
    writeonly = Parameter()

    def __init__(self):
        super(MockDevice, self).__init__()

    def _get_readonly(self):
        return 1

    def _set_writeonly(self, value):
        pass


class TestDevice(TestCase):

    def setUp(self):
        super(TestDevice, self).setUp()
        self.device = MockDevice()

    def test_accessor_functions(self):
        self.assertEqual(self.device.get_readonly().result(), 1)
        self.device.set_writeonly(0).join()

    def test_iterable(self):
        for param in self.device:
            self.assertTrue(param.name in ('readonly', 'writeonly'))

    def test_invalid_parameter(self):
        def query_invalid_param():
            self.device['foo']

        self.assertRaises(ParameterError, query_invalid_param)

    # def test_str(self):
    #     table = get_default_table(["Parameter", "Value"])
    #     table.border = False
    #     table.add_row(["readonly", "1"])
    #     table.add_row(["state", Device.NA])
    #     self.assertEqual(str(self.device), table.get_string())

    def test_context_manager(self):
        # This is just a functional test, we don't cover synchronization issues
        # here.
        with self.device as d:
            d.readonly
            d.writeonly = 2

    def test_parameter_lock_acquisition(self):
        with self.device['writeonly']:
            pass



"""
An experiment can be run multiple times. The base :py:class:`.Experiment`
takes care of proper logging structure.
"""

import os
import re
import logging
from logging import FileHandler, Formatter
from concert.async import async
from concert.storage import create_directory
from concert.coroutines.base import broadcast, inject


LOG = logging.getLogger(__name__)


class Acquisition(object):

    """
    An acquisition object connects data generator to consumers.

    .. py:attribute:: generator_caller

        a callable which returns a generator once called

    .. py:attribute:: consumer_callers

        a list of callables which return a coroutine once started

    """

    def __init__(self, name, generator_caller, consumer_callers=None):
        self.name = name
        self.generator = generator_caller
        self.consumers = [] if consumer_callers is None else consumer_callers

    def __call__(self):
        """Run the acquisition, i.e. connect the producer and consumer."""
        LOG.debug("Running acquisition '{}'".format(self))

        started = []
        for not_started in self.consumers:
            started.append(not_started())

        inject(self.generator(), broadcast(*started))

    def __repr__(self):
        return "Acquisition({})".format(self.name)


class Experiment(object):

    r"""
    Experiment base class. An experiment can be run multiple times
    with logging output saved on disk. The log from every
    :meth:`~.Experiment.run` is saved in the current experiment
    directory given by *directory_prefix*.

    .. py:attribute:: acquisitions

        A list of acquisitions this experiment is composed of

    .. py:attribute:: directory_prefix

       Directory prefix is either a formattable string in which case the
       at each experiment run a new directory given by the prefix and
       the current iteration is created. If the *directory_prefix* is a
       simple string then the individual experiment runs are stored in
       its subdirectories starting with scan\_ and suffixed by the run
       iteration.

    .. py:attribute:: log

        A logger to which a file handler will be attached in order to
        store the log output in the current directory

    .. py:attribute:: log_file_name

        Log file name used for storing logging information.

    """

    def __init__(self, acquisitions, directory_prefix, log=None,
                 log_file_name="experiment.log"):
        self.acquisitions = acquisitions
        self.directory_prefix = directory_prefix
        self.log = log
        self.log_file_name = log_file_name
        pattern = re.compile(".*\{.*\}.*")
        if pattern.match(self.directory_prefix) is None:
            self.directory_prefix = os.path.join(self.directory_prefix, "scan_{:>03}")
        self._file_stream = None
        self.iteration = 1

    def swap(self, first, second):
        """
        Swap acquisition *first* with *second*. If there are more occurences
        of either of them then the ones which are found first in the acquisitions
        list are swapped.
        """
        if first not in self.acquisitions or second not in self.acquisitions:
            raise ValueError("Both acquisitions must be part of the experiment")

        first_index = self.acquisitions.index(first)
        second_index = self.acquisitions.index(second)
        tmp = first
        self.acquisitions[first_index] = second
        self.acquisitions[second_index] = tmp

    def get_acquisition(self, name):
        """
        Get acquisition by its *name*. In case there are more like it, the first
        one is returned.
        """
        for acq in self.acquisitions:
            if acq.name == name:
                return acq
        raise ExperimentError("Acquisition with name `{}' not found".format(name))

    @property
    def directory(self):
        """Current directory for running the experiment."""
        return self.directory_prefix.format(self.iteration)

    def _create_stream_handler(self):
        """
        Create file stream handler for given scan to log information
        to the current directory.
        """
        path = os.path.join(self.directory, self.log_file_name)
        self._file_stream = FileHandler(path)
        self._file_stream.setLevel(logging.INFO)
        formatter = Formatter("[%(asctime)s] %(levelname)s: %(name)s: %(message)s")
        self._file_stream.setFormatter(formatter)
        self.log.addHandler(self._file_stream)

    def acquire(self):
        """
        Acquire data by running the acquisitions. This is the method which implements
        the data acquisition and should be overriden if more functionality is required,
        unlike :meth:`~.Experiment.run`.
        """
        for acq in self.acquisitions:
            acq()

    @async
    def run(self):
        """
        run()

        Create current directory, attach logging output to file and run the
        :meth:`~.base.Experiment.acquire`. After the run is complete the logging
        is cleaned up automatically. This method should *not* be overriden.
        """
        # Create directory for next scan
        while os.path.exists(self.directory):
            # Iterate until an unused directory has been found
            self.iteration += 1
        create_directory(self.directory)

        # Initiate new logger for this scan
        if self.log:
            self._create_stream_handler()

        try:
            self.acquire()
        finally:
            if self.log:
                self._file_stream.close()
                self.log.removeHandler(self._file_stream)


class ExperimentError(Exception):
    pass



import threading
import logging
from concert.base import Parameterizable


LOG = logging.getLogger(__name__)


class Device(Parameterizable):

    """
    A :class:`.Device` provides locked access to a real-world device.

    It implements the context protocol to provide locking::

        with device:
            # device is locked
            device.parameter = 1 * q.m
            ...

        # device is unlocked again
    """

    def __init__(self):
        # We have to create the lock early on because it will be accessed in
        # any add_parameter calls, especially those in the Parameterizable base
        # class
        self._lock = threading.Lock()
        super(Device, self).__init__()

    def __enter__(self):
        self._lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self._lock.release()



"""Connection protocols for network communication."""
import os
import logging
import socket
from threading import Lock


LOG = logging.getLogger(__name__)


class SocketConnection(object):

    """A two-way socket connection. *return_sequence* is a string appended
    after every command indicating the end of it, the default value
    is a newline (\\n).
    """

    def __init__(self, host, port, return_sequence="\n"):
        self._peer = (host, port)
        self._sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self._sock.settimeout(20)
        self._lock = Lock()
        self.return_sequence = return_sequence
        self._sock.connect(self._peer)

    def __del__(self):
        self._sock.close()

    def send(self, data):
        """
        Send *data* to the peer. The return sequence characters
        are appended to the data before it is sent.
        """
        LOG.debug('Sending {0}'.format(data))
        data += self.return_sequence
        self._sock.sendall(data.encode('ascii'))

    def recv(self):
        """
        Read data from the socket. The result is first stripped
        from the trailing return sequence characters and then returned.
        """
        try:
            result = self._sock.recv(1024)
            if result.endswith(self.return_sequence):
                # Strip the command-ending character
                result = result.rstrip(self.return_sequence)
            LOG.debug('Received {0}'.format(result))
            return result
        except socket.timeout:
            LOG.warn('Reading from %s:%i timed out' % self._peer)

    def execute(self, data):
        """Execute command and wait for response (thread safe)."""
        self._lock.acquire()
        try:
            self.send(data)
            result = self.recv()
        finally:
            self._lock.release()

        return result


def get_tango_device(uri, peer=None):
    """
    Get a Tango device by specifying its *uri*. If *peer* is given change the
    tango_host specifying which database to connect to. Format is host:port
    as a string.
    """
    import PyTango
    # TODO: check if there is a way to adjust the host in PyTango.
    if peer is not None:
        os.environ["TANGO_HOST"] = peer

    return PyTango.DeviceProxy(uri)



import numpy as np
import logging
from scipy.ndimage.filters import gaussian_filter
from concert.async import async, wait
from concert.quantities import q
from concert.measures import rotation_axis
from concert.optimization import halver, optimize_parameter
from concert.imageprocessing import center_of_mass, flat_correct
from concert.coroutines.base import coroutine


LOG = logging.getLogger(__name__)


def _pull_first(tuple_list):
        for tup in tuple_list:
            yield tup[0]


@async
def scan(param, feedback, minimum=None, maximum=None, intervals=64, convert=lambda x: x):
    """
    scan(param, feedback, minimum=None, maximum=None, intervals=64, convert=lambda x: x)

    Scan the parameter object in *intervals* steps between *minimum* and
    *maximum* and call *feedback* at each step. *feedback* must return a value
    that is evaluated at the parameter position. If *minimum* or *maximum* is
    ``None``, :attr:`.ParameterValue.lower` or :attr:`.ParameterValue.upper` is
    used.

    Set *convert* to a callable that transforms the parameter value prior to
    setting it.

    Returns a tuple *(x, y)* with parameter and feedback values.
    """
    minimum = minimum if minimum is not None else param.lower
    maximum = maximum if maximum is not None else param.upper

    xss = np.linspace(minimum, maximum, intervals)
    yss = []

    for xval in xss:
        param.set(convert(xval)).join()
        yss.append(feedback())

    return (xss, yss)


def scan_param_feedback(scan_param, feedback_param, minimum=None, maximum=None, intervals=64,
                        convert=lambda x: x):
    """
    Convenience function to scan one parameter and measure another.

    Scan the *scan_param* object and measure *feedback_param* at each of the
    *intervals* steps between *minimum* and *maximum*.

    Returns a tuple *(x, y)* with scanned parameter and measured values.
    """
    def feedback():
        return feedback_param.get().result()

    return scan(scan_param, feedback, minimum, maximum, intervals, convert)


def ascan(param_list, n_intervals, handler, initial_values=None):
    """
    For each of the *n_intervals* and for each of the *(parameter, start,
    stop)* tuples in *param_list*, calculate a set value from *(stop - start) /
    n_intervals* and set *parameter* to it::

        ascan([(motor['position'], 0 * q.mm, 2 * q.mm)], 5, handler)

    When all devices have reached the set point *handler* is called with a list
    of the parameters as its first argument.

    If *initial_values* is given, it must be a list with the same length as
    *devices* containing start values from where each device is scanned.
    """
    parameters = [param for param in _pull_first(param_list)]

    if initial_values:
        if len(param_list) != len(initial_values):
            raise ValueError("*initial_values* must match *parameter_list*")
    else:
        initial_values = []

        for param in parameters:
            if param.unit:
                initial_values.append(0 * param.unit)
            else:
                initial_values.append(0)

    initialized_params = list(map(lambda x: x[0] + (x[1],),
                                  zip(param_list, initial_values)))

    for i in range(n_intervals + 1):
        futures = []

        for param, start, stop, init in initialized_params:
            step = (stop - start) / n_intervals
            value = init + start + i * step
            futures.append(param.set(value))

        wait(futures)
        handler(parameters)


def dscan(parameter_list, n_intervals, handler):
    """
    For each of the *n_intervals* and for each of the *(parameter, start,
    stop)* tuples in *param_list*, calculate a set value from *(stop - start) /
    n_intervals* and set *parameter*.
    """
    initial_values = [param.get().result()
                      for param in _pull_first(parameter_list)]

    return ascan(parameter_list, n_intervals, handler, initial_values)


def focus(camera, motor, measure=np.std, opt_kwargs=None,
          plot_consumer=None, frame_consumer=None):
    """
    Focus *camera* by moving *motor*. *measure* is a callable that computes a
    scalar that has to be maximized from an image taken with *camera*.
    *opt_kwargs* are keyword arguments sent to the optimization algorithm.
    *plot_consumer* is fed with y values from the optimization and
    *frame_consumer* is fed with the incoming frames.

    This function is returning a future encapsulating the focusing event. Note,
    that the camera is stopped from recording as soon as the optimal position
    is found.
    """
    if opt_kwargs is None:
        opt_kwargs = {'initial_step': 0.5 * q.mm,
                      'epsilon': 1e-2 * q.mm}

    def get_measure():
        camera.trigger()
        frame = camera.grab()
        if frame_consumer:
            frame_consumer.send(frame)
        return - measure(frame)

    @coroutine
    def filter_optimization():
        """
        Filter the optimization's (x, y) subresults to only the y part,
        otherwise the the plot update is not lucid.
        """
        while True:
            tup = yield
            if plot_consumer:
                plot_consumer.send(tup[1])

    camera.trigger_mode = camera.trigger_modes.SOFTWARE
    camera.start_recording()
    f = optimize_parameter(motor['position'], get_measure, motor.position,
                           halver, alg_kwargs=opt_kwargs,
                           consumer=filter_optimization())
    f.add_done_callback(lambda unused: camera.stop_recording())
    return f


@async
def align_rotation_axis(camera, rotation_motor, x_motor=None, z_motor=None,
                        measure=rotation_axis, num_frames=10, absolute_eps=0.1 * q.deg,
                        max_iterations=5, flat=None, dark=None, frame_consumer=None):
    """
    align_rotation_axis(camera, rotation_motor, x_motor=None, z_motor=None,
    measure=rotation_axis, num_frames=10, absolute_eps=0.1 * q.deg, max_iterations=5,
    flat=None, dark=None, frame_consumer=None)

    Align rotation axis. *camera* is used to obtain frames, *rotation_motor*
    rotates the sample around the tomographic axis of rotation, *x_motor*
    turns the sample around x-axis, *z_motor* turns the sample around z-axis.
    *measure* provides axis of rotation angular misalignment data (a callable),
    *num_frames* defines how many frames are acquired and passed to the *measure*.
    *absolute_eps* is the threshold for stopping the procedure. If *max_iterations*
    is reached the procedure stops as well. *flat* and *dark* are the normalization
    frames applied on the acquired frames. *frame_consumer* is a coroutine which will
    receive the frames acquired at different sample positions.

    The procedure finishes when it finds the minimum angle between an
    ellipse extracted from the sample movement and respective axes or the
    found angle drops below *absolute_eps*. The axis of rotation after
    the procedure is (0,1,0), which is the direction perpendicular
    to the beam direction and the lateral direction.
    """
    if not x_motor and not z_motor:
        raise ValueError("At least one of the x, z motors must be given")

    def get_frames():
        frames = []
        for i in range(num_frames):
            rotation_motor.move(step).join()
            camera.trigger()
            frame = camera.grab().astype(np.float)
            if flat is not None:
                frame = flat_correct(frame, flat, dark=dark)
            if frame_consumer:
                frame_consumer.send(frame)
            frames.append(frame)

        return frames

    step = 2 * np.pi / num_frames * q.rad
    x_angle, z_angle, center = None, None, None

    if camera.state == 'recording':
        camera.stop_recording()
    camera['trigger_mode'].stash().join()
    camera.trigger_mode = camera.trigger_modes.SOFTWARE
    camera.start_recording()

    try:
        # Sometimes both z-directions need to be tried out because of the
        # projection ambiguity.
        z_direction = -1
        i = 0

        x_last = None
        z_last = None
        z_turn_counter = 0

        while True:
            x_angle, z_angle, center = measure(get_frames())

            x_better = True if z_motor is not None and\
                (x_last is None or np.abs(x_angle) < x_last) else False
            z_better = True if x_motor is not None and\
                (z_last is None or np.abs(z_angle) < z_last) else False

            if x_motor:
                if z_better:
                    z_turn_counter = 0
                elif z_turn_counter < 1:
                    # We might have rotated in the opposite direction because
                    # of the projection ambiguity. However, that must be picked up
                    # in the next iteration, so if the two consequent angles
                    # are worse than the minimum, we have the result.
                    z_better = True
                    z_direction = -z_direction
                    z_turn_counter += 1

            x_future, z_future = None, None
            if z_better and np.abs(z_angle) >= absolute_eps:
                x_future = x_motor.move(z_direction * z_angle)
            if x_better and np.abs(x_angle) >= absolute_eps:
                z_future = z_motor.move(x_angle)
            elif (np.abs(z_angle) < absolute_eps or not z_better):
                # The newly calculated angles are worse than the previous
                # ones or the absolute threshold has been reached,
                # stop iteration.
                break

            wait([future for future in [x_future, z_future] if future is not None])

            x_last = np.abs(x_angle)
            z_last = np.abs(z_angle)

            i += 1

            if i == max_iterations:
                # If we reached maximum iterations we consider it a failure
                # because the algorithm was not able to get to the desired
                # solution within the max_iterations limit.
                raise ProcessException("Maximum iterations reached")
    finally:
        camera.stop_recording()
        # No side effects
        camera['trigger_mode'].restore().join()

    # Return the last known ellipse fit
    return x_angle, z_angle, center


def find_beam(cam, xmotor, zmotor, pixelsize, xborder, zborder,
              xstep=None, zstep=None, thres=1000):
    """
    Scans the area defined by xborder and zborder for the beam until
    beam_visible returns True.
    Startpoint is the current motor-position if this position is inside the
    defined area else it start from the center of that area.
    It searches in a spiral around the startpoint.

    *cam* is the camera-device, *xmotor* the motor-device horizontally aligned
    to the image and *zmotor* the motor-device vertically aligned to the image.
    *pixelsize* determines the realworld size of an image pixels (scalar or
    2-element array-like, e.g. [4*q.um, 5*q.um]). *xborder* and *zborder*
    define the search area. Each constructed with a start- and an end-value
    (e.g. [-1.2*q.mm, 5.5*q.mm]).

    Optional arguments *xstep* and *zstep* define the length of one movement
    in the specific direction. Defaults are calculated from cam_img.shape and
    *pixelsize*.
    Optional argument *thres* will be past to beam_visible().
    """

    def beam_visible(img, thres):
        """
        Simple grayvalue-threshold that defines the beam to be visible in *img*.
        """
        return (img >= thres).any()

    def check(rel_move):
        """Move to new position and check if the beam is visible."""
        fut_0 = zmotor.move(rel_move[0]*q.um)
        fut_1 = xmotor.move(rel_move[1]*q.um)
        wait([fut_0, fut_1])
        cam.trigger()
        img = gaussian_filter(cam.grab().astype(np.float32), 40.0)
        img_shape[0] = img.shape
        bv = beam_visible(img, thres)
        return bv

    def spiral_scan():
        """
        Generator for returning rel. movement to the next position of the
        scanpath.
        """

        def trydir(dr):
            """
            Check if position in direction *dr* is unchecked or outside the
            search-area.
            Result:   2 - ok; 1 - outside; 0 - inside but already used
            """
            tmp = sp_pos + dir2rpos[dr]
            inside = (tmp >= 0).all() and (tmp < sp_shape).all()
            unused = (not scanned_points[tmp[0], tmp[1]]) if inside else False
            return 1 if not inside else 2 if unused else 0

        def next_move(pos, dr, rel_move=True):
            """
            Moves relativ in the direction *dr* (or to position *dr* if
            rel_move==False) and marks new position as scanned.
            Clips movement at border of the scanarea if needed.
            """
            old = np.array(pos).copy()
            if rel_move:
                pos += dir2rpos[dr]
            else:
                pos[0], pos[1] = dr[0], dr[1]
            scanned_points[pos[0], pos[1]] = True
            if (old == 0).any() or (old == sp_shape-1).any() or \
                    (pos == 0).any() or (pos == sp_shape-1).any():
                new_pos = np.array([zpos, xpos]) + step * (pos-[nz0, nx0])
                clip = np.append((new_pos < [zborder[0], xborder[0]]),
                                 (new_pos > [zborder[1], xborder[1]]))
                new_pos[0] = zborder[1] if clip[2] else \
                    zborder[0] if clip[0] else new_pos[0]
                new_pos[1] = xborder[1] if clip[3] else \
                    xborder[0] if clip[1] else new_pos[1]
                old_pos = np.array([zmotor.position.to(q.um.units).magnitude,
                                    xmotor.position.to(q.um.units).magnitude])
                return new_pos - old_pos
            else:
                if rel_move:
                    return step * dir2rpos[dr]
                else:
                    return step * (pos - old)

        def decide_dir():
            """
            Picks a direction to go and updates direction of search-rotation.
            """
            dr = np.array(map(trydir, range(4))) == 2
            r = np.arange(4)[dr][0]
            w0 = sp_pos-[nz0, nx0]
            w1 = w0 + dir2rpos[r]
            w = np.arctan2(w1[0], w1[1]) - np.arctan2(w0[0], w0[1])
            search_rot[0] = int((w % (2*np.pi)) < np.pi)
            return r

        xpos = xmotor.position.to(q.um.units).magnitude
        zpos = zmotor.position.to(q.um.units).magnitude
        search_rot = [0]
        rdir2try = [[1, 0, 3], [3, 0, 1]]
        dir2rpos = [[1, 0], [0, 1], [-1, 0], [0, -1]]
        step = img_shape[0] * \
            np.array([x.to(q.um.units).magnitude for x in ps])
        step[0] = step[0] if zstep is None else zstep.to(q.um.units).magnitude
        step[1] = step[1] if xstep is None else xstep.to(q.um.units).magnitude
        nz0 = int((zpos-zborder[0])/step[0]) + \
            ((zpos-zborder[0]) % step[0] > step[0]/2.)
        nz1 = int((zborder[1]-zpos)/step[0]) + \
            ((zborder[1]-zpos) % step[0] > step[0]/2.)
        nx0 = int((xpos-xborder[0])/step[1]) + \
            ((xpos-xborder[0]) % step[1] > step[1]/2.)
        nx1 = int((xborder[1]-xpos)/step[1]) + \
            ((xborder[1]-xpos) % step[1] > step[1]/2.)
        sp_shape = np.array([nz0+nz1+1, nx0+nx1+1])
        scanned_points = np.zeros(sp_shape, np.bool_)
        sp_pos = np.array([nz0, nx0])
        scanned_points[sp_pos[0], sp_pos[1]] = True

        while not scanned_points.all():
            skip = False
            try:
                sdir = decide_dir()
                yield next_move(sp_pos, sdir)
            except IndexError:
                skip = True

            while (not scanned_points.all()) and (not skip):
                sr = search_rot[0]
                for rd in rdir2try[sr]:
                    rd = (sdir + rd) % 4
                    c = trydir(rd)
                    if c == 2:
                        break
                    elif c == 1:
                        search_rot[0] = (search_rot[0]+1) % 2
                if c == 2:
                    sdir = rd
                    yield next_move(sp_pos, sdir)
                    continue
                break

            if not scanned_points.all():
                ind = np.indices(sp_shape)
                mask = np.logical_not(scanned_points)
                ind = np.array([ind[0][mask], ind[1][mask]])
                dist = ((ind-np.array([nz0, nx0]).reshape(2, 1))**2).sum(0)
                sort_pos = ind[:, np.argsort(dist)]
                n_pos = sort_pos[:, 0]
                r = next_move(sp_pos, n_pos, False)
                yield r

    units = (u for u in pixelsize.units if pixelsize.units[u] > 0.0)
    units = q.parse_expression(' * '.join(units))
    ps = np.tile(pixelsize.magnitude, 2)[:2] * units
    xborder = np.array([xborder[0].to(q.um.units).magnitude,
                        xborder[1].to(q.um.units).magnitude])
    zborder = np.array([zborder[0].to(q.um.units).magnitude,
                        zborder[1].to(q.um.units).magnitude])

    # startpoint
    img_shape = [0]
    xpos = xmotor.position.to(q.um.units).magnitude
    zpos = zmotor.position.to(q.um.units).magnitude
    outside = (xpos < xborder[0]) or (xpos > xborder[1]) or \
        (zpos < zborder[0]) or (zpos > zborder[1])
    rmov = (zborder.mean() - zpos, xborder.mean() - xpos) if outside \
        else (0*zpos, 0*xpos)
    if check(rmov):
        return True

    # scan
    for rmov in spiral_scan():
        if check(rmov):
            return True

    return False


def drift_to_beam(cam, xmotor, zmotor, pixelsize, tolerance=5,
                  max_iterations=100):
    """
    Moves the camera *cam* with motors *xmotor* and *zmotor* until the
    center of mass is nearer than *tolerance*-pixels to the center of the
    frame or *max_iterations* is reached.

    To convert pixelcoordinates to realworld-coordinates of the motors the
    *pixelsize* (scalar or 2-element array-like, e.g. [4*q.um, 5*q.um]) is
    needed.
    """
    def take_frame():
        cam.trigger()
        return gaussian_filter(cam.grab().astype(np.float32), 40.0)

    units = (u for u in pixelsize.units if pixelsize.units[u] > 0.0)
    units = q.parse_expression(' * '.join(units))
    ps = np.tile(pixelsize.magnitude, 2)[:2] * units

    img = take_frame()
    frm_center = (np.array(img.shape)-1)/2
    d = center_of_mass(img-img.min()) - frm_center

    iter_ = 0
    while ((d**2).sum() > tolerance**2) and (iter_ < max_iterations):
        fut_0 = zmotor.move(d[0]*ps[0])
        fut_1 = xmotor.move(d[1]*ps[1])
        wait([fut_0, fut_1])
        img = take_frame()
        if img.sum() == 0:
            LOG.debug("drift_to_beam: Frame is empty (sum == 0). " +
                      "Can't follow center of mass.")
            raise ProcessException("There is nothing to see! "
                                   "Can't follow the center of mass.")
        d = center_of_mass(img-img.min()) - frm_center
        iter_ += 1

    if iter_ < max_iterations:
        return True
    else:
        return False


@async
def center_to_beam(cam, xmotor, zmotor, pixelsize, xborder, zborder,
                   xstep=None, zstep=None, thres=1000, tolerance=5,
                   max_iterations=100):
    """
    Tries to center the camera *cam* to the beam by moving with the motors
    *xmotor* and *zmotor*. It starts by searching the beam inside the
    search-area defined by *xborder* and *zborder*. Argument *pixelsize* is
    needed to convert pixelcoordinates into realworld-coordinates of the
    motors.
    Exceptions are raised on fail.

    Optional arguments *xstep*, *zstep*, *thres*, *tolerance* and
    *max_iterations* are passed to the functions 'find_beam(...)' and
    'center2beam(...)'.
    """
    cam.trigger_mode = cam.trigger_modes.SOFTWARE
    cam.start_recording()

    try:
        with cam, xmotor, zmotor:
            if not find_beam(cam, xmotor, zmotor, pixelsize, xborder, zborder,
                             xstep, zstep, thres):
                message = "Unable to find the beam"
                LOG.debug('center_to_beam: '+message)
                raise ProcessException(message)
            else:
                LOG.debug('center_to_beam: switch to drift_to_beam')
                if not drift_to_beam(cam, xmotor, zmotor, pixelsize, tolerance,
                                     max_iterations):
                    message = "Maximum iterations reached"
                    LOG.debug('center_to_beam: '+message)
                    raise ProcessException(message)
    finally:
        cam.stop_recording()


class ProcessException(Exception):

    """
    Exception raised by a process when something goes wrong with the procedure
    it tries to accomplish, e.g. cannot focus, cannot align rotation axis, etc.

    """

    pass



import math
import inspect
import subprocess
import prettytable
from concert.devices.base import Device


def _get_param_description_table(device, max_width):
    field_names = ["Name", "Unit", "Description"]
    widths = [int(math.floor(f * max_width)) for f in (0.3, 0.2, 0.5)]
    table = get_default_table(field_names, widths)
    table.border = False
    table.header = True

    for param in device:
        units = param.unit if hasattr(param, 'unit') else None
        row = [param.name, str(units), str(param)]
        table.add_row(row)

    return table.get_string()


def _get_param_value_table(device):
    field_names = ["Name", "Value"]
    table = get_default_table(field_names)
    table.border = False
    table.header = False

    for param in device:
        table.add_row([param.name, str(param.get().result())])

    if hasattr(device, 'state'):
        table.add_row(['state', device.state])

    return table.get_string()


def _current_instances(instance_type):
    # Get the second frame to skip caller who called us
    frame = inspect.stack()[2]
    instances = frame[0].f_globals

    return ((name, obj) for (name, obj)
            in instances.items()
            if isinstance(obj, instance_type))


def get_default_table(field_names, widths=None):
    """Return a prettytable styled for use in the shell. *field_names* is a
    list of table header strings."""
    table = prettytable.PrettyTable(field_names)
    table.border = True
    table.hrules = prettytable.ALL
    table.vertical_char = ' '
    table.junction_char = '-'

    def left_align(name):
        # Try different ways of setting the alignment to support older versions
        # of prettytable.
        try:
            table.align[name] = 'l'
        except AttributeError:
            table.set_field_align(name, 'l')

    if widths:
        for name, width in zip(field_names, widths):
            left_align(name)
            table.max_width[name] = width
    else:
        for name in field_names:
            left_align(name)

    return table


def dstate():
    """Render device state in a table."""

    field_names = ["Name", "Parameters"]
    table = get_default_table(field_names)

    for name, device in _current_instances(Device):
        values = _get_param_value_table(device)
        table.add_row([name, values])

    print(table.get_string())


def ddoc():
    """Render device documentation."""
    from concert.devices.base import Device

    n_columns = int(subprocess.check_output(['stty', 'size']).split()[1]) - 8
    field_names = ["Name", "Description", "Parameters"]
    widths = [int(math.floor(f * n_columns)) for f in (0.1, 0.2, 0.7)]
    table = get_default_table(field_names, widths)

    for name, device in _current_instances(Device):
        doc = _get_param_description_table(device, widths[2] - 4)
        table.add_row([name, inspect.getdoc(device), doc])

    print(table.get_string())


def pdoc(hide_blacklisted=True):
    """Render process documentation."""
    black_listed = ('show', 'start', 'init', 'rm', 'log', 'edit', 'fetch')
    field_names = ["Name", "Description"]
    table = get_default_table(field_names)

    frame = inspect.stack()[1]
    instances = frame[0].f_globals

    for name, obj in instances.items():
        if not name.startswith('_') and inspect.isfunction(obj):
            if hide_blacklisted and name not in black_listed:
                table.add_row([name, inspect.getdoc(obj)])

    print(table.get_string())


def code_of(func):
    """Show implementation of *func*."""
    source = inspect.getsource(func)

    try:
        from pygments import highlight
        from pygments.lexers import PythonLexer
        from pygments.formatters import TerminalFormatter

        print(highlight(source, PythonLexer(), TerminalFormatter()))
    except ImportError:
        print(source)



"""
Image processing module for manipulating image data, e.g. filtered
backprojection, flat field correction and other operations on images.
"""

import numpy as np
import logging
from scipy import ndimage
from scipy.signal import fftconvolve
from concert.quantities import q


LOG = logging.getLogger(__name__)


def flat_correct(radio, flat, dark=None):
    """
    Flat field correction of a radiograph *radio* with *flat* field.
    If *dark* field is supplied it is taken into account as well.
    """
    return radio / flat if dark is None else (radio - dark) / (flat - dark)


def ramp_filter(width):
    """Get a 1D ramp filter for filtering sinogram rows."""
    base = np.arange(-width / 2, width / 2)

    return np.fft.fftshift(np.abs(base)) * 2.0 / width


def needle_tips(images):
    """Get sample tips from images."""
    tips = []

    for image in images:
        tip = _get_ellipse_point(image)
        if tip is not None:
            tips.append(tip)

    if len(tips) == 0:
        raise ValueError("No sample tip points found.")

    return tips


def _segment(image):
    """
    Segment a flat corrected *image* into a sample and a background.

    Assume normally distributed noise, take the full width at
    1/1000 of the maximum and make it a threshold for finding
    the sample. The sample must be highly absorbing. The baseline
    for the background is taken from the top row of the image.
    """
    thr = 0.5
    image[image < thr] = 0
    image[image >= thr] = 1
    image = 1 - image
    image = ndimage.binary_fill_holes(np.cast[np.bool](image))

    # Close the image to get rid of noise-caused fuzzy segmentation.
    return ndimage.binary_closing(image) + image


def _get_boundary_coordinates(coordinates, max_val):
    """Return coordinates which reside on image edges."""
    return [coor for coor in coordinates if coor % max_val == 0]


def _is_corner_point(point, shape):
    """Test if the *point* lies in one of the image corners."""
    return (point[1] == 0 or point[1] == shape[1] - 1) and\
        (point[0] == 0 or point[0] == shape[0] - 1)


def _get_intersection_points(image):
    """Get *image* edges and sample intersection points. The *image* is
    a segmented binary image."""
    y_ind, x_ind = np.where(image != 0)
    x_low = x_ind[np.where(y_ind == 0)]
    x_high = x_ind[np.where(y_ind == image.shape[0] - 1)]
    y_low = y_ind[np.where(x_ind == 0)]
    y_high = y_ind[np.where(x_ind == image.shape[1] - 1)]

    points = []
    if len(x_low) != 0:
        points.append((0, x_low[0]))
        if x_low[-1] != x_low[0]:
            points.append((0, x_low[-1]))
    if len(x_high) != 0:
        points.append((image.shape[0] - 1, x_high[0]))
        if x_high[-1] != x_high[0]:
            points.append((image.shape[0] - 1, x_high[-1]))
    if len(y_low) != 0:
        points.append((y_low[0], 0))
        if y_low[-1] != y_low[0]:
            points.append((y_low[-1], 0))
    if len(y_high) != 0:
        points.append((y_high[0], image.shape[1] - 1))
        if y_high[-1] != y_high[0]:
            points.append((y_high[-1], image.shape[1] - 1))

    if len(points) > 2:
        # The sample is big and besides intersection points it fills some
        # corners of the image.
        res = []
        for point in points:
            if not _is_corner_point(point, image.shape):
                res.append(point)
        points = res

    return points


def _get_axis_intersection(p_1, p_2, shape):
    """Get intersections of a vector perpendicular to a vector defined by
    *p_1* and *p_2* and image edges defined by image *shape*."""
    # First check if the center lies on an edge
    if p_1[0] == p_2[0]:
        return [(p_1[0], (p_1[1] + p_2[1]) / 2)]
    elif p_1[1] == p_2[1]:
        return [((p_1[0] + p_2[0]) / 2.0, p_1[1])]

    p_x = (p_1[1] + p_2[1]) / 2.0
    p_y = (p_1[0] + p_2[0]) / 2.0
    v_y = p_1[0] - p_2[0]
    v_x = p_2[1] - p_1[1]
    height, width = shape[0] - 1, shape[1] - 1

    left = p_y - v_x * p_x / v_y, 0
    right = p_y + v_x * (width - p_x) / v_y, width
    bottom = 0, p_x - v_y * p_y / v_x
    top = height, p_x + v_y * (height - p_y) / v_x

    res = set([left, right, bottom, top])
    # Filter intersections which are out of the image bounding box.
    res = [x for x in res if 0 <= x[0] <= height and 0 <= x[1] <= width]

    return res


def _get_sample_tip(image):
    """Extract sample tip from the labeled *image*."""
    # First check if the sample tip is in the FOV. If not, there are no
    # objects or the sample splits the image in half.
    if ndimage.label(image.max() - image)[1] != 1:
        # No sample found in this image. It is either completely out of the
        # FOV or it splits the image in half, thus the tip is out of the FOV.
        return None

    p_1, p_2 = _get_intersection_points(image)
    intersections = _get_axis_intersection(p_1, p_2, image.shape)

    if len(intersections) == 0:
        return None

    p_inter = None

    # Get the intersection where a sample is present.
    for intersection in intersections:
        if image[intersection] > 0:
            p_inter = intersection
            break

    if p_inter is None:
        return None

    y_ind, x_ind = np.where(image > 0)

    # Calculate distances from the intersection point.
    distances = np.sqrt((y_ind - p_inter[0]) ** 2 + (x_ind - p_inter[1]) ** 2)

    # Most distant points are candidates for the sample tip.
    indices = np.where(distances == distances.max())[0]

    tips = [(y_ind[i], x_ind[i]) for i in indices]
    tip_center = center_of_points(tips)

    # Now find the intersection of the object and line going through it
    # in the direction to the tip.
    x_direction = tip_center[1] - p_inter[1]
    if x_direction == 0:
        # The line is parallel with y_ind axis
        x_ind = np.ones(image.shape[1], dtype=np.int64) * p_inter[1]
        y_ind = np.arange(image.shape[0])
    else:
        y_direction = tip_center[0] - p_inter[0]
        x_ind = np.arange(image.shape[1])
        y_ind = np.cast[np.int](np.round(p_inter[0] + y_direction * (
                                         x_ind - p_inter[1]) / x_direction))

    # Cut values going beyond image boundaries.
    below = np.where(y_ind < image.shape[0])[0]
    y_ind = y_ind[below]
    x_ind = x_ind[below]
    above = np.where(0 <= y_ind)[0]
    y_ind = y_ind[above].astype(np.int)
    x_ind = x_ind[above].astype(np.int)

    # Drop indices at which there is no object in the image.
    nonzero = np.where(image[y_ind, x_ind] > 0)[0]
    x_ind = x_ind[nonzero]
    y_ind = y_ind[nonzero]

    distances = np.sqrt((y_ind - p_inter[0]) ** 2 + (x_ind - p_inter[1]) ** 2)
    i = distances.argmax()

    return y_ind[i], x_ind[i]


def _get_sample(image):
    """Segment the *image* to a sample and a background."""
    labels = _get_regions(image)
    labels = _get_boundary_regions(labels)
    # Cut the edges by one pixel because hole filling does not affect them.
    return _get_biggest_region(labels)[1:-1, 1:-1]


def _get_ellipse_point(image):
    """Extract ellipse points from the sample in *image*."""
    labels = _get_sample(_segment(image))

    return _get_sample_tip(labels)


def _get_biggest_region(image):
    """Get the region with the biggest area in the *image*."""
    labels, features = ndimage.label(image)
    sizes = ndimage.sum(image, labels, range(features + 1))
    max_feature = np.argmax(sizes)
    labels[labels != max_feature] = 0

    return labels


def _get_boundary_regions(labels):
    """Extract regions from *labels* which appear on image boundaries only."""
    features = np.max(labels)

    for i in range(features + 1):
        y_ind, x_ind = np.where(labels == i)
        if len(x_ind) != 0 and len(y_ind) != 0 and x_ind.min() != 0 and\
                x_ind.max() != labels.shape[1] - 1 and y_ind.min() != 0 and\
                y_ind.max() != labels.shape[0] - 1:
            labels[labels == i] = 0

    return labels


def _get_regions(image):
    """Extract regions from the binary *image* which are not too small."""
    labels, features = ndimage.label(image)
    sizes = ndimage.sum(image, labels, range(features + 1))

    # Remove small regions.
    mask = sizes < 100
    remove_indices = mask[labels]
    labels[remove_indices] = 0

    return labels


def center_of_points(points):
    """
    Find a simplified center of mass withouth point-weighing
    from a set of *points*.
    """
    y_ind, x_ind = zip(*points)

    c_y = float(np.sum(y_ind)) / len(points)
    c_x = float(np.sum(x_ind)) / len(points)

    return c_y, c_x


def center_of_mass(frame):
    """Calculates the center of mass of the whole frame wheighted by value."""

    frm_shape = np.array(frame.shape)
    total = frame.sum()
    if total == 0:
        return np.array([-1, -1])
    else:
        y = (frame.sum(1)*np.arange(frm_shape[0])).sum() / total
        x = (frame.sum(0)*np.arange(frm_shape[1])).sum() / total
        return np.array([y, x])


def compute_rotation_axis(first_projection, last_projection):
    """
    Compute the tomographic rotation axis based on cross-correlation technique.
    *first_projection* is the projection at 0 deg, *last_projection* is the projection
    at 180 deg. It is strongly advised to provide flat-corrected projections.
    """
    width = first_projection.shape[1]
    # The rotation by 180 deg flips the image horizontally, in order
    # to do cross-correlation by convolution we must also flip it
    # vertically, so the image is transposed and we can apply convolution
    # which will act as cross-correlation
    convolved = fftconvolve(first_projection, last_projection[::-1, :], mode='same')
    center = np.unravel_index(convolved.argmax(), convolved.shape)[1]

    return (width / 2.0 + center) / 2 * q.px



"""
.. exception:: KillException

    Exception that may be thrown during the execution of an :func:`.async`
    decorated function. The function may run cleanup code.

.. function:: async

    A decorator for functions which are executed asynchronously.

.. function:: threaded

    Threaded execution of a function *func*.

"""
import functools
import traceback
import concert.config
from concurrent.futures import ThreadPoolExecutor, Future

try:
    import Queue as queue
except ImportError:
    import queue


# Provide nicely formatted exceptions if possible
_colored_exceptions = True
try:
    from IPython.core.ultratb import AutoFormattedTB
except ImportError:
    _colored_exceptions = False


# Patch futures so that they provide a join() and kill() method
def _join(self, _timeout=None):
    self.result()
    return self


def _kill(self, exception=None, block=True, timeout=None):
    pass


Future.join = _join
Future.kill = _kill


class NoFuture(Future):

    def __init__(self, result):
        super(NoFuture, self).__init__()
        self.set_result(result)


def no_async(func):
    @functools.wraps(func)
    def _inner(*args, **kwargs):
        future = NoFuture(None)
        try:
            result = func(*args, **kwargs)
            future.set_result(result)
        except Exception as e:
            if _colored_exceptions:
                AutoFormattedTB(mode='Verbose')()
            else:
                traceback.print_exc()
            future.set_exception(e)

        return future

    return _inner


try:
    if concert.config.DISABLE_GEVENT:
        raise ImportError

    import gevent
    import gevent.monkey
    import gevent.threadpool

    gevent.monkey.patch_all()

    # XXX: we have to import threading after patching
    import threading

    HAVE_GEVENT = True
    KillException = gevent.GreenletExit
    threadpool = gevent.threadpool.ThreadPool(4)

    class GreenletFuture(gevent.Greenlet):

        """A Future interface based on top of a Greenlet.

        This class provides the :class:`concurrent.futures.Future` interface on
        top of a Greenlet.
        """

        def __init__(self, func, args, kwargs):
            super(GreenletFuture, self).__init__()
            self.func = func
            self.args = args
            self.kwargs = kwargs
            self.saved_exception = None
            self._cancelled = False
            self._running = False

        def _run(self, *args, **kwargs):
            try:
                self._running = True
                value = self.func(*self.args, **self.kwargs)
                self._running = False

                # Force starting at least a bit of the greenlet
                gevent.sleep(0)
                return value
            except Exception as exception:
                self.saved_exception = exception

        def join(self, timeout=None):
            super(GreenletFuture, self).join(timeout)

            if self.saved_exception:
                raise self.saved_exception

        def cancel(self):
            self._cancelled = True
            self.kill()
            return True

        def cancelled(self):
            return self._cancelled

        def running(self):
            return self._running

        def done(self):
            return self.ready()

        def result(self, timeout=None):
            value = self.get(timeout=timeout)

            if self.saved_exception:
                raise self.saved_exception

            return value

        def add_done_callback(self, callback):
            self.link(callback)

    def async(func):
        @functools.wraps(func)
        def _inner(*args, **kwargs):
            g = GreenletFuture(func, args, kwargs)
            g.start()
            return g

        return _inner

    def threaded(func):
        @functools.wraps(func)
        def _inner(*args, **kwargs):
            result = threadpool.spawn(func, *args, **kwargs)
            return result

        return _inner

except ImportError:
    import threading
    HAVE_GEVENT = False

    # Module-wide executor
    EXECUTOR = ThreadPoolExecutor(max_workers=128)

    # This is a stub exception that will never be raised.
    class KillException(Exception):
        pass

    def async(func):
        if concert.config.DISABLE_ASYNC:
            return no_async(func)
        else:
            @functools.wraps(func)
            def _inner(*args, **kwargs):
                return EXECUTOR.submit(func, *args, **kwargs)

            return _inner

    def threaded(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            """Execute in a separate thread."""
            thread = threading.Thread(target=func, args=args, kwargs=kwargs)
            thread.daemon = True
            thread.start()
            return thread

        return wrapper


def wait(futures):
    """Wait for the list of *futures* to finish and raise exceptions if
    happened."""
    for future in futures:
        future.join()


class Dispatcher(object):

    """Core dispatcher"""

    def __init__(self):
        self._subscribers = {}
        self._messages = queue.Queue()
        self._event_queues = {}
        self._lock = threading.Lock()

        server = threading.Thread(target=self._serve)
        server.daemon = True
        server.start()

    def subscribe(self, sender, message, handler):
        """Subscribe to a message sent by sender.

        When message is sent by sender, handler is called with sender as the
        only argument.

        """
        tsm = sender, message
        if tsm in self._subscribers:
            self._subscribers[tsm].add(handler)
        else:
            self._subscribers[tsm] = set([handler])

    def unsubscribe(self, sender, message, handler):
        """Remove *handler* from the subscribers to *(sender, message)*."""
        tsm = sender, message
        if tsm in self._subscribers:
            self._subscribers[tsm].remove(handler)

    def send(self, sender, message):
        """Send message from sender."""
        self._messages.put((sender, message))

    def _serve(self):
        while True:
            tsm = self._messages.get()
            sender, message = tsm

            if tsm in self._subscribers:
                for callback in self._subscribers[tsm]:
                    callback(sender)

            if tsm in self._event_queues:
                self._event_queues[tsm].notify_and_clear()

            self._messages.task_done()


dispatcher = Dispatcher()



"""Port, IO Device."""
from concert.base import AccessorNotImplementedError
from concert.devices.base import Device


class IO(Device):

    """The IO device consists of ports which can be readable, writable or
    both.
    """

    def __init__(self):
        super(IO, self).__init__()
        self._ports = []

    @property
    def ports(self):
        """Port IDs used by :meth:`.read_port` and :meth:`.write_port`"""
        return self._ports

    def _check(self, port):
        """Check if *port* is a part of the device."""
        if port not in self._ports:
            raise IODeviceError("Port `{}' not found".format(port))

    def read_port(self, port):
        """Read a *port*."""
        self._check(port)
        return self._read_port(port)

    def write_port(self, port, value):
        """Write a *value* to the *port*."""
        self._check(port)
        self._write_port(port, value)

    def _read_port(self, port):
        """Implementation of reading a *port* from the device."""
        raise AccessorNotImplementedError

    def _write_port(self, port, value):
        """Implementation of writing a *value* to a *port* on the device."""
        raise AccessorNotImplementedError


class IODeviceError(Exception):

    """Specific exception thrown by IO devices."""

    pass



class Command(object):
    """Command class for the CLI script"""

    def __init__(self, name, opts):
        """
        Command objects are loaded at run-time and injected into Concert's
        command parser.

        *name* denotes the name of the sub-command parser, e.g. "mv" for the
        MoveCommand. *opts* must be an argparse-compatible dictionary
        of command options.
        """
        self.name = name
        self.opts = opts

    def run(self, *args, **kwargs):
        """Run the command"""
        raise NotImplementedError


class Bunch(object):
    """Encapsulate a list or dictionary to provide attribute-like access.

    Common use cases look like this::

        d = {'foo': 123, 'bar': 'baz'}
        b = Bunch(d)
        print(b.foo)
        >>> 123

        l = ['foo', 'bar']
        b = Bunch(l)
        print(b.foo)
        >>> 'foo'
    """
    def __init__(self, values):
        if isinstance(values, list):
            values = dict(zip(values, values))
        self.__dict__.update(values)


def memoize(func):
    """
    Memoize the result of *func*.

    Remember the result of *func* depending on its arguments. Note, that this
    requires that the function is free from any side effects, e.g. returns the
    same value given the same arguments.
    """
    memo = {}

    def wrapper(*args):
        if args in memo:
            return memo[args]

        result = func(*args)
        memo[args] = result
        return result

    return wrapper



import scipy
from scipy.ndimage import fourier
from concert.tests import assert_almost_equal, suppressed_logging
from concert.quantities import q
from concert.processes import focus
from concert.devices.cameras.dummy import Base as DummyCameraBase
from concert.devices.motors.dummy import LinearMotor


MIN_POSITION = 0 * q.mm
MAX_POSITION = 100 * q.mm
FOCUS_POSITION = 35 * q.mm


class BlurringCamera(DummyCameraBase):

    def __init__(self, motor):
        super(BlurringCamera, self).__init__()
        self._original = scipy.misc.lena()
        self.motor = motor

    def _grab_real(self):
        sigma = abs((self.motor.position - FOCUS_POSITION).magnitude)
        return fourier.fourier_gaussian(self._original, sigma)


@suppressed_logging
def test_focusing():
    motor = LinearMotor()
    motor.position = 40. * q.mm
    camera = BlurringCamera(motor)
    focus(camera, motor).join()
    assert_almost_equal(motor.position, FOCUS_POSITION, 1e-2)



import time
from concert.tests import TestCase
from concert.base import transition, StateError, TransitionNotAllowed, State
from concert.quantities import q
from concert.async import async
from concert.devices.base import Device


STOP_VELOCITY = 0 * q.mm / q.s
MOVE_VELOCITY = 1 * q.mm / q.s


class SomeDevice(Device):

    state = State(default='standby')

    def __init__(self):
        super(SomeDevice, self).__init__()
        self.velocity = STOP_VELOCITY

    @transition(source='standby', target='moving')
    def start_moving(self, velocity):
        self.velocity = velocity

    @transition(source='*', target='standby')
    def stop_moving(self):
        self.velocity = STOP_VELOCITY

    @async
    @transition(source='standby', target='standby', immediate='moving')
    def move_some_time(self, velocity, duration):
        self.velocity = velocity
        time.sleep(duration)
        self.velocity = STOP_VELOCITY

    @transition(source=['standby', 'moving'], target='standby')
    def stop_no_matter_what(self):
        self.velocity = STOP_VELOCITY

    def actual_state(self):
        return 'standby' if not self.velocity else 'moving'

    @transition(source='*', target=['standby', 'moving'], check=actual_state)
    def set_velocity(self, velocity):
        self.velocity = velocity

    @transition(source='*', target=['ok', 'error'])
    def make_error(self):
        raise StateError('error')

    @transition(source='error', target='standby')
    def reset(self):
        pass


class BaseDevice(Device):

    state = State(default='standby')

    def __init__(self):
        super(BaseDevice, self).__init__()

    @transition(source='standby', target='in-base')
    def switch_base(self):
        pass

    @transition(source='*', target='standby')
    def reset(self):
        pass


class DerivedDevice(BaseDevice):
    state = State(default='standby')

    def __init__(self):
        super(DerivedDevice, self).__init__()

    @transition(source='standby', target='in-derived')
    def switch_derived(self):
        pass


class TestStateMachine(TestCase):

    def setUp(self):
        super(TestStateMachine, self).setUp()
        self.device = SomeDevice()

    def test_state_per_device(self):
        d1 = SomeDevice()
        d2 = SomeDevice()

        d2.start_moving(MOVE_VELOCITY)
        self.assertNotEqual(d1.state, d2.state)

    def test_defaults(self):
        self.device.state == 'standby'

    def test_valid_transition(self):
        self.device.start_moving(MOVE_VELOCITY)
        self.assertEqual(self.device.state, 'moving')

        self.device.stop_moving()
        self.assertEqual(self.device.state, 'standby')

    def test_async_transition(self):
        future = self.device.move_some_time(MOVE_VELOCITY, 0.01)
        future.join()
        self.assertEqual(self.device.state, 'standby')

    def test_multiple_source_states(self):
        self.device.stop_no_matter_what()
        self.assertEqual(self.device.state, 'standby')

        self.device.start_moving(MOVE_VELOCITY)
        self.device.stop_no_matter_what()

        self.assertEqual(self.device.state, 'standby')

    def test_multiple_target_states(self):
        self.device.set_velocity(MOVE_VELOCITY)
        self.assertEqual(self.device.state, 'moving')

        self.device.set_velocity(STOP_VELOCITY)
        self.assertEqual(self.device.state, 'standby')

    def test_inheritance(self):
        device = DerivedDevice()

        device.switch_base()
        self.assertEqual(device.state, 'in-base')

        with self.assertRaises(TransitionNotAllowed):
            device.switch_base()

        with self.assertRaises(TransitionNotAllowed):
            device.switch_derived()

        device.reset()
        self.assertEqual(device.state, 'standby')

        device.switch_derived()
        self.assertEqual(device.state, 'in-derived')

        with self.assertRaises(TransitionNotAllowed):
            device.switch_derived()

        with self.assertRaises(TransitionNotAllowed):
            device.switch_base()

        device.reset()
        self.assertEqual(device.state, 'standby')

    def test_errors(self):
        with self.assertRaises(StateError):
            self.device.make_error()

        self.assertEqual(self.device.state, 'error')
        self.device.reset()
        self.assertEqual(self.device.state, 'standby')

    def test_state_setting(self):
        with self.assertRaises(AttributeError):
            self.device.state = 'foo'



"""Base scales module for implementing scales."""

from concert.devices.base import Device
from concert.base import Quantity, AccessorNotImplementedError
from concert.quantities import q
from concert.async import async


class WeightError(Exception):

    """Class for weighing errors."""

    pass


class Scales(Device):

    """Base scales class."""
    weight = Quantity(q.g, help="Weighted mass")

    def __init__(self):
        super(Scales, self).__init__()

    def _get_weight(self):
        raise AccessorNotImplementedError


class TarableScales(Scales):

    """Scales which can be tared."""

    def __init__(self):
        super(TarableScales, self).__init__()

    @async
    def tare(self):
        """Tare the scales."""
        self._tare()

    def _tare(self):
        raise AccessorNotImplementedError



from __future__ import absolute_import
import threading
import numpy as np

try:
    from gi.repository import Ufo
    import ufo.numpy
except ImportError as e:
    print(str(e))

from concert.coroutines.base import coroutine


class PluginManager(object):

    """Plugin manager that initializes new tasks."""

    def __init__(self):
        self._wrapped = Ufo.PluginManager()

    def get_task(self, name, **kwargs):
        """
        Create a new task from plugin *name* and initialize with *kwargs*.
        """
        task = self._wrapped.get_task(name)
        task.set_properties(**kwargs)
        return task


class InjectProcess(object):

    """Process to inject NumPy data into a UFO processing graph.

    :class:`InjectProcess` can also be used as a context manager, in which
    case it will call :meth:`~.InjectProcess.start` on entering the manager
    and :meth:`.wait` on exiting it.

    *graph* must either be a Ufo.TaskGraph or a Ufo.TaskNode object.  If it is
    a graph the input tasks will be connected to the roots, otherwise a new
    graph will be created.
    """

    def __init__(self, graph, get_output=False):
        self.output_task = None
        self._started = False

        if isinstance(graph, Ufo.TaskGraph):
            self.graph = graph
            roots = self.graph.get_roots()
        elif isinstance(graph, Ufo.TaskNode):
            self.graph = Ufo.TaskGraph()
            roots = [graph]
        else:
            msg = 'graph is neither Ufo.TaskGraph nor Ufo.TaskNode'
            raise ValueError(msg)

        # Initialize inputs
        self.input_tasks = {}
        self.ufo_buffers = {}
        for root in roots:
            self.input_tasks[root] = []
            self.ufo_buffers[root] = []
            num_inputs = root.get_num_inputs()
            for i in range(num_inputs):
                self.input_tasks[root].append(Ufo.InputTask())
                self.ufo_buffers[root].append(None)
                self.graph.connect_nodes_full(self.input_tasks[root][i], root, i)

        if get_output:
            self.output_task = Ufo.OutputTask()
            leaves = self.graph.get_leaves()
            self.graph.connect_nodes(leaves[0], self.output_task)

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.wait()
        return True

    @coroutine
    def __call__(self, consumer):
        """Co-routine compatible consumer."""
        if not self._started:
            self.start()

        while True:
            item = yield
            self.insert(item)
            consumer.send(self.result())

    def start(self):
        """
        Run the processing in a new thread.

        Use :meth:`.push` to insert data into the processing chaing and
        :meth:`.wait` to wait until processing has finished."""
        def run_scheduler():
            sched = Ufo.Scheduler()
            sched.run(self.graph)

        self.thread = threading.Thread(target=run_scheduler)
        self.thread.start()
        if not self._started:
            self._started = True

    def insert(self, array, node=None, index=0):
        """
        Insert *array* into the *node*'s *index* input.

        .. note:: *array* must be a NumPy compatible array.
        """
        if not node:
            if len(self.input_tasks) > 1:
                raise ValueError('input_node cannot be None for graphs with more inputs')
            else:
                node = self.input_tasks.keys()[0]
        if self.ufo_buffers[node][index] is None:
            self.ufo_buffers[node][index] = ufo.numpy.fromarray(array.astype(np.float32))
        else:
            self.ufo_buffers[node][index] = self.input_tasks[node][index].get_input_buffer()
            ufo.numpy.fromarray_inplace(self.ufo_buffers[node][index], array.astype(np.float32))

        self.input_tasks[node][index].release_input_buffer(self.ufo_buffers[node][index])

    def result(self):
        if self.output_task:
            buf = self.output_task.get_output_buffer()
            result = np.copy(ufo.numpy.asarray(buf))
            self.output_task.release_output_buffer(buf)
            return result

    def wait(self):
        """Wait until processing has finished."""
        for input_tasks in self.input_tasks.values():
            for input_task in input_tasks:
                input_task.stop()
        self.thread.join()


class Backproject(InjectProcess):

    """
    Coroutine to reconstruct slices from sinograms using filtered
    backprojection.

    *axis_pos* specifies the center of rotation in pixels within the sinogram.
    If not specified, the center of the image is assumed to be the center of
    rotation.
    """

    def __init__(self, axis_pos=None):
        self.pm = PluginManager()
        self.fft = self.pm.get_task('fft', dimensions=1)
        self.ifft = self.pm.get_task('ifft', dimensions=1)
        self.fltr = self.pm.get_task('filter')
        self.backprojector = self.pm.get_task('backproject')
        self.crop = self.pm.get_task('region-of-interest', x=0, y=0)

        if axis_pos:
            self.backprojector.props.axis_pos = axis_pos

        graph = Ufo.TaskGraph()
        graph.connect_nodes(self.fft, self.fltr)
        graph.connect_nodes(self.fltr, self.ifft)
        graph.connect_nodes(self.ifft, self.backprojector)
        graph.connect_nodes(self.backprojector, self.crop)

        super(Backproject, self).__init__(graph, get_output=True)

    @property
    def axis_position(self):
        return self.backprojector.props.axis_pos

    @axis_position.setter
    def axis_position(self, position):
        self.backprojector.props.axis_pos = position

    @coroutine
    def __call__(self, consumer):
        """Get a sinogram, do filtered backprojection and send it to *consumer*."""
        if not self._started:
            self.start()

        while True:
            sinogram = yield
            self.insert(sinogram)
            self.crop.props.width = sinogram.shape[1]
            self.crop.props.height = sinogram.shape[1]
            slice = self.result()

            consumer.send(slice)


class FlatCorrectedBackproject(InjectProcess):

    """
    Coroutine to reconstruct slices from sinograms using filtered
    backprojection. The data are first flat-field corrected and then
    backprojected. All the inputs must be of type unsigned int 16.

    *flat_row* is a row of a flat field, *dark_row* is a row of the dark field.
    The rows must correspond to the sinogram which is being backprojected.
    *axis_pos* specifies the center of rotation in pixels within the sinogram.
    If not specified, the center of the image is assumed to be the center of
    rotation.
    """

    def __init__(self, flat_row, dark_row, axis_pos=None):
        self.pm = PluginManager()
        self.sino_correction = self.pm.get_task('sino-correction')
        self.fft = self.pm.get_task('fft', dimensions=1)
        self.ifft = self.pm.get_task('ifft', dimensions=1)
        self.fltr = self.pm.get_task('filter')
        self.backprojector = self.pm.get_task('backproject')
        self.crop = self.pm.get_task('region-of-interest', x=0, y=0)

        if axis_pos:
            self.backprojector.props.axis_pos = axis_pos

        graph = Ufo.TaskGraph()
        graph.connect_nodes(self.sino_correction, self.fft)
        graph.connect_nodes(self.fft, self.fltr)
        graph.connect_nodes(self.fltr, self.ifft)
        graph.connect_nodes(self.ifft, self.backprojector)
        graph.connect_nodes(self.backprojector, self.crop)

        super(FlatCorrectedBackproject, self).__init__(graph, get_output=True)

        self.flat_row = flat_row
        self.dark_row = dark_row

    @property
    def axis_position(self):
        return self.backprojector.props.axis_pos

    @axis_position.setter
    def axis_position(self, position):
        self.backprojector.props.axis_pos = position

    @property
    def dark_row(self):
        return self._dark_row

    @dark_row.setter
    def dark_row(self, row):
        self._dark_row = row.astype(np.float32)

    @property
    def flat_row(self):
        return self._flat_row

    @flat_row.setter
    def flat_row(self, row):
        self._flat_row = row.astype(np.float32)

    @coroutine
    def __call__(self, consumer):
        """Get a sinogram, do filtered backprojection and send it to *consumer*."""
        if not self._started:
            self.start()

        while True:
            sinogram = yield
            self.insert(sinogram.astype(np.float32), node=self.sino_correction, index=0)
            self.insert(self.dark_row, node=self.sino_correction, index=1)
            self.insert(self.flat_row, node=self.sino_correction, index=2)
            self.crop.props.width = sinogram.shape[1]
            self.crop.props.height = sinogram.shape[1]
            slice = self.result()

            consumer.send(slice)



"""
Cameras supported by the libuca library.
"""
import functools
import logging
import time
import numpy as np
from concert.quantities import q
from concert.base import Parameter, Quantity
from concert.helpers import Bunch
from concert.devices.cameras import base


LOG = logging.getLogger(__name__)


def _new_setter_wrapper(name, unit=None):
    def _wrapper(instance, value):
        if unit:
            value = value.to(unit)

        try:
            dic = {name: value.magnitude}
        except AttributeError:
            dic = {name: value}

        instance.uca.set_properties(**dic)

    return _wrapper


def _new_getter_wrapper(name, unit=None):
    def _wrapper(instance):
        value = instance.uca.get_property(name)

        if unit:
            return value * unit

        return value

    return _wrapper


def _translate_gerror(func):
    from gi.repository import GLib

    @functools.wraps(func)
    def _wrapper(*args, **kwargs):
        try:
            return func(*args, **kwargs)
        except GLib.GError as ge:
            raise base.CameraError(str(ge))

    return _wrapper


class Camera(base.Camera):

    """libuca-based camera.

    All properties that are exported by the underlying camera are also visible.
    """

    def __init__(self, name):
        """
        Create a new libuca camera.

        The *name* is passed to the uca plugin manager.

        :raises CameraError: In case camera *name* does not exist.
        """

        super(Camera, self).__init__()

        from gi.repository import GObject, Uca

        self._manager = Uca.PluginManager()

        try:
            self.uca = self._manager.get_camerav(name, [])
        except:
            raise base.CameraError("`{0}' is not a valid camera".format(name))

        units = {
            Uca.Unit.METER: q.m,
            Uca.Unit.SECOND: q.s,
            Uca.Unit.DEGREE_CELSIUS: q.celsius,
            Uca.Unit.COUNT: q.dimensionless,
            Uca.Unit.PIXEL: q.pixel,
        }

        parameters = {}

        for prop in self.uca.props:
            if prop.name == 'trigger-mode' or prop.name == 'frames-per-second':
                continue

            getter, setter, unit = None, None, None

            uca_unit = self.uca.get_unit(prop.name)

            if uca_unit in units:
                unit = units[uca_unit]

            if prop.flags & GObject.ParamFlags.READABLE:
                getter = _new_getter_wrapper(prop.name, unit)

            if prop.flags & GObject.ParamFlags.WRITABLE:
                setter = _new_setter_wrapper(prop.name, unit)

            name = prop.name.replace('-', '_')

            if uca_unit in units:
                parameters[name] = Quantity(unit, fget=getter, fset=setter, help=prop.blurb)
            else:
                parameters[name] = Parameter(fget=getter, fset=setter, help=prop.blurb)

        if parameters:
            self.install_parameters(parameters)

        class _Dummy(object):
            pass

        setattr(self.uca, 'enum_values', _Dummy())

        def get_enum_bunch(enum):
            enum_map = {}

            for key, value in enum.__enum_values__.items():
                name = value.value_nick.upper().replace('-', '_')
                enum_map[name] = key

            return Bunch(enum_map)

        for prop in self.uca.props:
            if hasattr(prop, 'enum_class'):
                setattr(self.uca.enum_values, prop.name.replace('-', '_'),
                        get_enum_bunch(prop.default_value))

        self._uca_get_frame_rate = _new_getter_wrapper('frames-per-second')
        self._uca_set_frame_rate = _new_setter_wrapper('frames-per-second')

        # Invert the uca trigger mode dict in order to return concert values
        trigger_dict = self.uca.enum_values.trigger_mode.__dict__
        self._uca_to_concert_trigger = {v: k for k, v in trigger_dict.items()}
        self._uca_get_trigger = _new_getter_wrapper('trigger-mode')
        self._uca_set_trigger = _new_setter_wrapper('trigger-mode')

        self._record_shape = None
        self._record_dtype = None

    def _get_frame_rate(self):
        return self._uca_get_frame_rate(self) / q.s

    def _set_frame_rate(self, frame_rate):
        self._uca_set_frame_rate(self, frame_rate * q.s)

    def _get_trigger_mode(self):
        uca_trigger = self._uca_get_trigger(self)
        return self._uca_to_concert_trigger[uca_trigger]

    def _set_trigger_mode(self, mode):
        uca_value = getattr(self.uca.enum_values.trigger_mode, mode)
        self._uca_set_trigger(self, uca_value)

    @_translate_gerror
    def _record_real(self):
        self._record_shape = self.roi_height.magnitude, self.roi_width.magnitude
        self._record_dtype = np.uint16 if self.sensor_bitdepth.magnitude > 8 else np.uint8
        self.uca.start_recording()

    @_translate_gerror
    def _stop_real(self):
        self.uca.stop_recording()

    @_translate_gerror
    def _trigger_real(self):
        self.uca.trigger()

    @_translate_gerror
    def _grab_real(self):
        array = np.empty(self._record_shape, dtype=self._record_dtype)
        data = array.__array_interface__['data'][0]

        if self.uca.grab(data):
            return array

        return None


class Pco(Camera):

    """Pco camera implemented by libuca."""

    def __init__(self):
        super(Pco, self).__init__('pco')

    def stream(self, consumer):
        """stream frames to the *consumer*."""
        try:
            self.acquire_mode = self.uca.enum_values.acquire_mode.AUTO
            self.storage_mode = self.uca.enum_values.storage_mode.RECORDER
            self.record_mode = self.uca.enum_values.record_mode.RING_BUFFER
        except:
            pass

        return super(Pco, self).stream(consumer)


class PCO4000(Pco):

    """PCO.4000 camera implementation."""

    _ALLOWED_DURING_RECORDING = ['trigger', '_trigger_real', '_last_grab_time',
                                 'grab', '_grab_real', '_record_shape', '_record_dtype',
                                 'uca', 'stop_recording']

    def __init__(self):
        self._lock_access = False
        self._last_grab_time = None
        super(PCO4000, self).__init__()

    def start_recording(self):
        super(PCO4000, self).start_recording()
        self._lock_access = True

    def stop_recording(self):
        self._lock_access = False
        super(PCO4000, self).stop_recording()

    def _grab_real(self):
        # For the PCO.4000 there must be a delay of at least one second
        # between consecutive grabs, otherwise it crashes. We provide
        # appropriate timeout here.
        current = time.time()
        if self._last_grab_time and current - self._last_grab_time < 1:
            time.sleep(1 - current + self._last_grab_time)
        result = super(PCO4000, self)._grab_real()
        self._last_grab_time = time.time()

        return result

    def __getattribute__(self, name):
        if object.__getattribute__(self, '_lock_access') and name \
           not in PCO4000._ALLOWED_DURING_RECORDING:
            raise AttributeError('PCO.4000 is inaccessible during recording')
        return object.__getattribute__(self, name)


class Dimax(Pco, base.BufferedMixin):

    """A pco.dimax camera implementation."""

    def __init__(self):
        super(Dimax, self).__init__()

    def _readout_real(self, num_frames=None):
        """Readout *num_frames* frames."""
        if num_frames is None:
            num_frames = self.recorded_frames

        if not 0 < num_frames <= self.recorded_frames:
            raise base.CameraError("Number of frames {} ".format(num_frames) +
                                   "must be more than zero and less than the recorded " +
                                   "number of frames {}".format(self.recorded_frames))

        try:
            self.uca.start_readout()

            for i in xrange(num_frames):
                yield self.grab()
        finally:
            self.uca.stop_readout()



from concert.tests import TestCase
from concert.devices.io.base import IODeviceError
from concert.devices.io.dummy import IO


class TestIO(TestCase):

    def setUp(self):
        self.io = IO(port_value=0)
        self.port = 0

    def test_read(self):
        self.assertEquals(0, self.io.read_port(self.port))

    def test_write(self):
        value = 1
        self.io.write_port(self.port, value)
        self.assertEquals(value, self.io.read_port(self.port))

    def test_non_existent_read(self):
        self.assertRaises(IODeviceError, self.io.read_port, 1)

    def test_non_existent_write(self):
        self.assertRaises(IODeviceError, self.io.write_port, 1, 0)



from concert.tests import TestCase
from concert.coroutines.base import coroutine
from concert.quantities import q
from concert.devices.cameras.dummy import Camera, BufferedCamera


class TestDummyCamera(TestCase):

    def setUp(self):
        super(TestDummyCamera, self).setUp()
        self.camera = Camera()

    def test_grab(self):
        frame = self.camera.grab()
        self.assertIsNotNone(frame)

    def test_trigger_mode(self):
        self.camera.trigger_mode = self.camera.trigger_modes.EXTERNAL
        self.assertEqual(self.camera.trigger_mode, 'EXTERNAL')

    def test_roi(self):
        self.roi_x0 = self.roi_y0 = self.roi_width = self.roi_height = 10 * q.dimensionless
        self.assertEqual(self.roi_x0, 10 * q.dimensionless)
        self.assertEqual(self.roi_y0, 10 * q.dimensionless)
        self.assertEqual(self.roi_width, 10 * q.dimensionless)
        self.assertEqual(self.roi_height, 10 * q.dimensionless)

    def test_has_sensor_dims(self):
        self.assertTrue(hasattr(self.camera, "sensor_pixel_width"))
        self.assertTrue(hasattr(self.camera, "sensor_pixel_height"))

    def test_buffered_camera(self):
        camera = BufferedCamera()
        for i, item in enumerate(camera.readout_buffer()):
            pass
        self.assertEqual(i, 2)

    def test_context_manager(self):
        camera = Camera()

        with camera.recording():
            self.assertEqual(camera.state, 'recording')
            f = camera.grab()

        self.assertEqual(camera.state, 'standby')

    def test_stream(self):
        @coroutine
        def check():
            while True:
                yield
                check.ok = True
                self.camera.stop_recording()
        check.ok = False

        self.camera.stream(check()).join()
        self.assertTrue(check.ok)



"""
.. data:: DISABLE_ASYNC

    Disable asynchronous execution by returning a dummy future which is not
    executed synchronusly.

.. data:: DISABLE_GEVENT

    Turn of gevent support and fall back to ThreadPoolExecutor approach.
"""

DISABLE_ASYNC = False
DISABLE_GEVENT = True



"""
Test experiments. Logging is disabled, so just check the directory and log
files creation.
"""
import shutil
import os
import tempfile
import numpy as np
from concert.quantities import q
from concert.coroutines.base import coroutine, inject
from concert.experiments.base import Acquisition, Experiment, ExperimentError
from concert.experiments.imaging import (Experiment as ImagingExperiment,
                                         tomo_angular_step, tomo_max_speed,
                                         tomo_projections_number, frames)
from concert.devices.cameras.dummy import Camera
from concert.tests import TestCase, suppressed_logging, assert_almost_equal


@suppressed_logging
def test_tomo_angular_step():
    truth = np.arctan(2.0 / 100) * q.rad
    assert_almost_equal(truth, tomo_angular_step(100 * q.px))


@suppressed_logging
def test_projections_number():
    width = 100
    truth = int(np.ceil(np.pi / np.arctan(2.0 / width) * q.rad))
    assert truth == tomo_projections_number(width * q.px)


@suppressed_logging
def test_tomo_max_speed():
    width = 100
    frame_rate = 100 / q.s
    truth = np.arctan(2.0 / width) * q.rad * frame_rate
    assert_almost_equal(truth, tomo_max_speed(width * q.px, frame_rate))


@suppressed_logging
def test_frames():
    @coroutine
    def count():
        while True:
            yield
            count.i += 1
    count.i = 0
    inject(frames(5, Camera()), count())
    assert count.i == 5


class TestExperimentBase(TestCase):

    def setUp(self):
        super(TestExperimentBase, self).setUp()
        self.base_directory = tempfile.mkdtemp()
        self.item = None

    def tearDown(self):
        shutil.rmtree(self.base_directory)

    @coroutine
    def consume(self):
        while True:
            self.item = yield


class TestExperiment(TestExperimentBase):

    def setUp(self):
        super(TestExperiment, self).setUp()
        self.item = None
        self.start = 0
        self.foo = Acquisition("foo", self.produce, consumer_callers=[self.consume])
        self.bar = Acquisition("bar", self.produce, consumer_callers=[self.consume])
        self.acquisitions = [self.foo, self.bar]
        self.experiment = Experiment(self.acquisitions, self.base_directory)

    def produce(self, start=0):
        for i in range(self.start, self.start + 2):
            yield i

    def test_run(self):
        self.experiment.run().join()
        self.assertTrue(os.path.exists(self.experiment.directory))
        self.assertEqual(self.item, 1)

        self.start = 2
        self.experiment.run().join()
        self.assertTrue(os.path.exists(self.experiment.directory))
        self.assertEqual(self.item, 3)

    def test_swap(self):
        self.experiment.swap(self.foo, self.bar)
        self.assertEqual(self.acquisitions[0], self.bar)
        self.assertEqual(self.acquisitions[1], self.foo)

    def test_get_by_name(self):
        self.assertEqual(self.foo, self.experiment.get_acquisition('foo'))
        self.assertRaises(ExperimentError, self.experiment.get_acquisition, 'non-existing')


class TestImagingExperiment(TestExperimentBase):

    def setUp(self):
        super(TestImagingExperiment, self).setUp()
        self.foo = Acquisition("foo", self.produce, consumer_callers=[self.consume])
        self.experiment = ImagingExperiment([self.foo], os.path.join(self.base_directory,
                                            "scan_{:>03}"))

    def produce(self):
        for i in range(2):
            yield np.ones((2, 2)) * (i + 1)

    def test_run(self):
        self.experiment.run().join()
        directory = os.path.join(self.experiment.directory, self.foo.name)
        self.assertEqual(sorted(os.listdir(directory)), ['frame_00000.tif', 'frame_00001.tif'])



# -*- coding: utf-8 -*-
"""Core module Parameters"""
import numpy as np
import logging
import six
import collections
import functools
import inspect
import types
from concert.helpers import memoize
from concert.async import async, wait


LOG = logging.getLogger(__name__)


def identity(x):
    return x


class TransitionNotAllowed(Exception):
    pass


class StateError(Exception):

    """Raised in check functions of state transitions of devices."""

    def __init__(self, error_state, msg=None):
        self.state = error_state
        self.msg = msg if msg else "Got into `{}'".format(error_state)

    def __str__(self):
        return self.msg


class UnitError(ValueError):

    """Raised when an operation is passed value with an incompatible unit."""
    pass


class LimitError(ValueError):

    """Raised when an operation is passed a value that exceeds a limit."""
    pass


class SoftLimitError(LimitError):

    """Raised when a soft limit is hit on the device."""
    pass


class HardLimitError(StateError):

    """Raised when device goes into hardlimit error state."""

    def __init__(self, error_state='hard-limit', msg=None):
        super(HardLimitError, self).__init__(error_state, msg)


class ParameterError(Exception):

    """Raised when a parameter is accessed that does not exists."""

    def __init__(self, parameter):
        msg = "`{0}' is not a parameter".format(parameter)
        super(ParameterError, self).__init__(msg)


class AccessorNotImplementedError(NotImplementedError):

    """Raised when a setter or getter is not implemented."""


class ReadAccessError(Exception):

    """Raised when user tries to change a parameter that cannot be written."""

    def __init__(self, parameter):
        LOG.warn("Invalid read access on {0}".format(parameter))
        msg = "parameter `{0}' cannot be read".format(parameter)
        super(ReadAccessError, self).__init__(msg)


class WriteAccessError(Exception):

    """Raised when user tries to read a parameter that cannot be read."""

    def __init__(self, parameter):
        LOG.warn("Invalid write access on {0}".format(parameter))
        msg = "parameter `{0}' cannot be written".format(parameter)
        super(WriteAccessError, self).__init__(msg)


class LockError(Exception):

    """Raised when parameter is locked."""

    pass


class MultiContext(object):

    """Multi context manager to be used in a Python `with` management.

    For example, to use multiple axes safely in one process, all you have to do
    is ::

        with MultiContext([axis1, axis2]):
            axis1.set_position()
            axis2.set_position()

    Original code by João S. O. Bueno licensed under CC-BY-3.0.
    """

    def __init__(self, *args):
        if (len(args) == 1 and
               (hasattr(args[0], "__len__") or
                hasattr(args[0], "__iter__"))):
            self.objs = list(args[0])
        else:
            self.objs = args

    def __enter__(self):
        return tuple(obj.__enter__() for obj in self.objs)

    def __exit__(self, type_, value, traceback):
        return all([obj.__exit__(type_, value, traceback)
                    for obj in self.objs])


class State(object):

    """
    Finite state machine.

    Use this on a class, to keep some sort of known state. In order to enforce
    restrictions, you would decorate methods on the class with
    :func:`.transition`::

        class SomeObject(object):

            state = State(default='standby')

            @transition(source='*', target='moving')
            def move(self):
                pass

    Accessing the state variable will return the current state value, i.e.::

        obj = SomeObject()
        assert obj.state == 'standby'

    The state cannot be set explicitly by::

        obj.state = 'some_state'

    but the object needs to provide methods which transition out of
    states, the same holds for transitioning out of error states.
    """

    def __init__(self, default=None):
        self.default = default

    def __get__(self, instance, owner):
        return self._value(instance)

    def __set__(self, instance, value):
        raise AttributeError('State cannot be set')

    def _value(self, instance):
        if not hasattr(instance, '_state_value'):
            setattr(instance, '_state_value', self.default)

        return getattr(instance, '_state_value')


def transition(source='*', target=None, immediate=None, check=None):
    """
    Decorates a method that triggers state transitions.

    *source* denotes the source state that must be present at the time of
    invoking the decorated method. *target* is the state that the state object
    will be after successful completion of the method or a list of possible
    target states. *immediate* is an optional state that will be set during
    execution of the method.

    *check* is a callable that will be called to determine the actual state in
    case *target* is a list of possible target states. Hence, when *check* is
    called it must return one of those target states.
    """
    def wrapped(func):
        transitions = collections.defaultdict(list)

        sources = [source] if isinstance(source, str) else source
        targets = [target] if isinstance(target, str) else target

        if immediate:
            sources.append(immediate)
            targets.append(immediate)

        for s in sources:
            transitions[s] = targets

        def _value(instance):
            if not hasattr(instance, '_state_value'):
                setattr(instance, '_state_value', instance.state)
            return instance.state

        def try_transition(target, instance, *args, **kwargs):
            current = _value(instance)
            succ = transitions.get(current, transitions.get('*', None))

            if not succ:
                msg = "Cannot transition from `{}' to `{}'".format(current, target)
                raise TransitionNotAllowed(msg)

        @functools.wraps(func)
        def call_func(instance, *args, **kwargs):
            current = _value(instance)

            if current not in sources and '*' not in sources:
                msg = "`{}' not in `{}'".format(source, sources)
                raise TransitionNotAllowed(msg)

            if immediate:
                # Since it was listed by the user, the transition must exist so no check required.
                setattr(instance, '_state_value', immediate)

            try:
                # If there is an edge to the desired transition from source or immediate
                # we execute the actual function
                try_transition(target, instance)

                if isinstance(func, types.MethodType):
                    result = func(*args, **kwargs)
                else:
                    result = func(instance, *args, **kwargs)

                # The final state can come from the device itself (more possible target states)
                final = getattr(instance, check.__name__)() if isinstance(target, list) else target
                # We check the actual state after the function execution before we do the final
                # transition
                try_transition(final, instance)
                setattr(instance, '_state_value', final)
            except StateError as error:
                setattr(instance, '_state_value', error.state)
                raise error

            return result

        return call_func

    return wrapped


class Parameter(object):

    """A parameter with getter and setter.

    Parameters are similar to normal Python properties and can additionally
    trigger state transitions. If *fget* or *fset* is not given, you must
    implement the accessor functions named `_set_name` and `_get_name`::

        from concert.base import Parameter, State

        class SomeClass(object):

            state = State(default='standby')

            def actual(self):
                return 'moving'

            param = Parameter(transition=transition(source='standby',
                                                    target=['standby', 'moving'],
                                                    check=actual))

            def _set_param(self, value):
                pass

            def _get_param(self):
                pass

    When a :class:`.Parameter` is attached to a class, you can modify it by
    accessing its associated :class:`.ParameterValue` with a dictionary
    access::

        obj = SomeClass()
        print(obj['param'])
    """

    def __init__(self, fget=None, fset=None, data=None, transition=None, help=None):
        """
        *fget* is a callable that is called when reading the parameter. *fset*
        is called when the parameter is written to.

        *data* is passed to the state transition function.

        *transition* is a :func:`.transition` that changes states when a value
        is written to the parameter.

        *help* is a string describing the parameter in more detail.
        """
        self.name = None
        self.fget = fget
        self.fset = fset
        self.data_args = (data,) if data is not None else ()
        self.transition = transition
        self.decorated = None
        self.help = help

    @memoize
    def setter_name(self):
        if self.fset:
            return self.fset.__name__

        return '_set_' + self.name

    @memoize
    def getter_name(self):
        if self.fget:
            return self.fget.__name__

        return '_get_' + self.name

    def __repr__(self):
        return str(self.help)

    def __get__(self, instance, owner):
        # If we would just call self.fset(value) we would call the method
        # defined in the base class. This is a hack (?) to call the function on
        # the instance where we actually want the function to be called.

        try:
            if self.fget:
                value = self.fget(instance, *self.data_args)
            else:
                value = getattr(instance, self.getter_name())(*self.data_args)

            return value
        except AccessorNotImplementedError:
            raise ReadAccessError(self.name)

    def __set__(self, instance, value):
        def log_access(what):
            """Log access."""
            msg = "{}: {}: {}='{}'"
            name = instance.__class__.__name__
            LOG.info(msg.format(name, what, self.name, value))

        log_access('try')

        if instance[self.name].locked:
            raise LockError("Parameter `{}' is locked for writing".format(self))

        if self.fset:
            self.fset(instance, value, *self.data_args)
        else:
            func = getattr(instance, '_set_' + self.name)

            if self.transition and not hasattr(func, '_is_transitioned'):
                func = self.transition(func)
                func._is_transitioned = True
                setattr(instance, '_set_' + self.name, func)

            try:
                if isinstance(func, types.MethodType):
                    func(value, *self.data_args)
                else:
                    func(instance, value, *self.data_args)
            except AccessorNotImplementedError:
                raise WriteAccessError(self.name)

        log_access('set')


class Quantity(Parameter):

    """A :class:`.Parameter` associated with a unit."""

    def __init__(self, unit, fget=None, fset=None, lower=None, upper=None,
                 data=None, transition=None, help=None):
        """
        *fget*, *fset*, *data*, *transition* and *help* are identical to the
        :class:`.Parameter` constructor arguments.

        *unit* is a Pint quantity. *lower* and *upper* denote soft limits
        between the :class:`.Quantity` values can lie.
        """
        super(Quantity, self).__init__(fget=fget, fset=fset, data=data,
                                       transition=transition, help=help)
        self.unit = unit

        self.upper = upper if upper is not None else float('Inf')
        self.lower = lower if lower is not None else -float('Inf')

        if upper is None:
            self.upper = self.upper * unit

        if lower is None:
            self.lower = self.lower * unit

    def is_compatible(self, value):
        try:
            self.unit + value
            return True
        except ValueError:
            return False

    def convert(self, value):
        return value.to(self.unit)

    def __get__(self, instance, owner):
        # If we would just call self.fset(value) we would call the method
        # defined in the base class. This is a hack (?) to call the function on
        # the instance where we actually want the function to be called.
        value = super(Quantity, self).__get__(instance, owner)

        return self.convert(value)

    def __set__(self, instance, value):
        if not self.is_compatible(value):
            msg = "{} of {} can only receive values of unit {} but got {}"
            raise UnitError(
                msg.format(self.name, type(instance), self.unit, value))

        lower = instance[self.name].lower
        upper = instance[self.name].upper

        def leq(a, b):
            """check a <= b"""
            if not hasattr(a, 'shape') or len(a.shape) == 0:
                return a <= b
            else:
                # Vector data type
                valid_a = np.invert(np.isnan(a.magnitude))
                valid_b = np.invert(np.isnan(b.magnitude))
                valid = valid_a & valid_b
                test_a = a[valid]
                test_b = b[valid]
                return np.all(test_a.to_base_units().magnitude <= test_b.to_base_units().magnitude)

        if not leq(lower, value) or not leq(value, upper):
            msg = "{} is out of range [{}, {}]"
            raise SoftLimitError(msg.format(value, lower, upper))

        converted = self.convert(value)
        super(Quantity, self).__set__(instance, converted)


def quantity(unit=None, lower=None, upper=None, data=None, transition=None, help=None):
    """
    Decorator for read-only quantity functions.

    Device authors can add additional read-only quantities to a specific device
    by applying this decorator to a function::

        class SomeDevice(Device):
            @quantity(unit=1*q.mm, lower=0*q.mm, upper=2*q.mm)
            def position(self):
                pass

    The arguments correspond to those of :class:`.Quantity`.
    """

    def wrapper(func):
        # Get help info from doc string if no explicit help was passed.
        doc = help if help else inspect.getdoc(func)

        return Quantity(fget=func, unit=unit, lower=lower, upper=upper,
                        data=data, transition=transition, help=doc)

    return wrapper


class ParameterValue(object):

    """Value object of a :class:`.Parameter`."""

    def __init__(self, instance, parameter):
        self._lock = None
        self._locked = False
        self._instance = instance
        self._parameter = parameter
        self._saved = []

    def __enter__(self):
        if self._lock is not None:
            self._lock.acquire()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self._lock is not None:
            self._lock.release()

    def __lt__(self, other):
        return self._parameter.name < other._parameter.name

    def __repr__(self):
        return self.info_table.get_string()

    @property
    def info_table(self):
        from concert.session.utils import get_default_table
        locked = "yes" if self.locked else "no"
        table = get_default_table(["attribute", "value"])
        table.header = False
        table.border = False
        table.add_row(["info", self._parameter.help])
        table.add_row(["locked", locked])
        return table

    @property
    def name(self):
        return self._parameter.name

    @property
    def data(self):
        return self._parameter.data

    @async
    def get(self):
        return getattr(self._instance, self.name)

    @async
    def set(self, value):
        setattr(self._instance, self.name, value)

    @async
    def stash(self):
        """Save the current value internally on a growing stack.

        If the parameter is writable the current value is saved on a stack and
        to be later retrieved with :meth:`.ParameterValue.restore`.
        """
        self._saved.append(self.get().result())

    def restore(self):
        """Restore the last value saved with :meth:`.ParameterValue.stash`.

        If the parameter can only be read or no value has been saved, this
        operation does nothing.
        """
        if self._saved:
            val = self._saved.pop()
            return self.set(val)

    @property
    def locked(self):
        """Return True if the parameter is locked for writing."""
        return self._locked

    def lock(self, permanent=False):
        """Lock parameter for writing. If *permament* is True the parameter
        cannot be unlocked anymore.
        """
        def unlock_not_allowed():
            raise LockError("Parameter `{}' cannot be unlocked".format(self))

        self._locked = True
        if permanent:
            self.unlock = unlock_not_allowed

    def unlock(self):
        """Unlock parameter for writing."""
        self._locked = False


class QuantityValue(ParameterValue):

    def __init__(self, instance, quantity):
        super(QuantityValue, self).__init__(instance, quantity)
        self._lower = quantity.lower
        self._upper = quantity.upper
        self._limits_locked = False

    def lock_limits(self, permanent=False):
        """Lock limits, if *permanent* is True the limits cannot be unlocker anymore."""
        def unlock_not_allowed():
            raise LockError('Limits are locked permanently')

        self._limits_locked = True
        if permanent:
            self.unlock_limits = unlock_not_allowed

    def unlock_limits(self):
        """Unlock limits."""
        self._limits_locked = False

    @property
    def lower(self):
        return self._lower

    @lower.setter
    def lower(self, value):
        if self._limits_locked:
            raise LockError('lower limit locked')
        else:
            self._lower = value

    @property
    def upper(self):
        return self._upper

    @upper.setter
    def upper(self, value):
        if self._limits_locked:
            raise LockError('upper limit locked')
        else:
            self._upper = value

    @property
    def info_table(self):
        table = super(QuantityValue, self).info_table
        table.add_row(["lower", self.lower])
        table.add_row(["upper", self.upper])
        return table

    @property
    def unit(self):
        return self._parameter.unit


class MetaParameterizable(type):

    def __init__(cls, name, bases, dic):
        super(MetaParameterizable, cls).__init__(name, bases, dic)

        def get_base_parameter_names():
            for base in bases:
                if hasattr(base, '_parameter_names'):
                    return getattr(base, '_parameter_names')
            return {}

        if not hasattr(cls, '_parameter_names'):
            setattr(cls, '_parameter_names', get_base_parameter_names())

        for attr_name, attr_type in dic.items():
            if isinstance(attr_type, Parameter):
                attr_type.name = attr_name
                cls._parameter_names[(cls, attr_name)] = attr_type


class Parameterizable(six.with_metaclass(MetaParameterizable, object)):

    """
    Collection of parameters.

    For each class of type :class:`.Parameterizable`, :class:`.Parameter` can
    be set as class attributes ::

        class Device(Parameterizable):

            def get_something(self):
                return 'something'

            something = Parameter(get_something)

    There is a simple :class:`.Parameter` and a parameter which models a
    physical quantity :class:`.Quantity`.

    A :class:`.Parameterizable` is iterable and returns its parameters of type
    :class:`.ParameterValue` or its subclasses ::

        for param in device:
            print("name={}".format(param.name))

    To access a single name parameter object, you can use the ``[]`` operator::

        param = device['position']
        print param.is_readable()

    If the parameter name does not exist, a :class:`.ParameterError` is raised.

    Each parameter value is accessible as a property. If a device has a
    position it can be read and written with::

        param.position = 0 * q.mm
        print param.position
    """

    def __init__(self):
        if not hasattr(self, '_params'):
            self._params = {}

        for (cls, name), parameter in self._parameter_names.items():
            if not isinstance(self, cls):
                continue

            self._install_parameter(name, parameter)

    def __str__(self):
        from concert.session.utils import get_default_table

        table = get_default_table(["Parameter", "Value"])
        table.border = False

        for param in self:
            table.add_row([param.name, str(param.get().result())])

        # If self would be Stateful, I'd feel better ...
        if hasattr(self, 'state'):
            table.add_row(['state', self.state])

        return table.get_string(sortby="Parameter")

    def __repr__(self):
        return '\n'.join([super(Parameterizable, self).__repr__(),
                         str(self)])

    def __iter__(self):
        for param in sorted(self._params.values()):
            yield param

    def __getitem__(self, param):
        if param not in self._params:
            raise ParameterError(param)

        return self._params[param]

    def install_parameters(self, params):
        """Install parameters at run-time.

        *params* is a dictionary mapping parameter names to :class:`.Parameter`
        objects.
        """
        for name, param in params.items():
            param.name = name
            self._install_parameter(name, param)

            # Install param as a property, so that it can be accessed via
            # object-dot notation.
            setattr(self.__class__, name, param)

    def _install_parameter(self, name, param):
        if param.__class__ == Quantity:
            value = QuantityValue(self, param)
        elif param.__class__ == Parameter:
            value = ParameterValue(self, param)

        self._params[name] = value

        def setter_not_implemented(value, *args):
            raise AccessorNotImplementedError

        def getter_not_implemented(*args):
            raise AccessorNotImplementedError

        setattr(self, 'set_' + name, value.set)
        setattr(self, 'get_' + name, value.get)

        if not hasattr(self, '_set_' + name):
            setattr(self, '_set_' + name, setter_not_implemented)

        if not hasattr(self, '_get_' + name):
            setattr(self, '_get_' + name, getter_not_implemented)

    @async
    def stash(self):
        """
        Save all writable parameters that can be restored with
        :meth:`.Parameterizable.restore`.

        The values are stored on a stacked, hence subsequent saved states can
        be restored one by one.
        """
        wait((param.stash() for param in self))

    @async
    def restore(self):
        """Restore all parameters saved with :meth:`.Parameterizable.stash`."""
        wait((param.restore() for param in self))

    def lock(self, permanent=False):
        """Lock all the parameters for writing. If *permanent* is True, the
        parameters cannot be unlocked anymore.
        """
        for param in self:
            param.lock(permanent=permanent)

    def unlock(self):
        """Unlock all the parameters for writing."""
        for param in self:
            param.unlock()


class Process(Parameterizable):

    """Base process."""

    def __init__(self, params=None):
        super(Process, self).__init__(params)

    @async
    def run(self):
        """run()

        Run the process. The result depends on the actual process.
        """
        raise NotImplementedError



"""Aerotech controllers IO boards implementation."""
from concert.networking.aerotech import Connection
from concert.devices.io import base


def _make_ids(index, num_items):
    return zip([index] * num_items, range(num_items))


class HLe(base.IO):

    """Aerotech HLe controller IO board."""

    def __init__(self, host, port=8001):
        super(HLe, self).__init__()
        self._connection = Connection(host, port)
        # Three groups, first 0-5, the rest 0-7
        self._ports = _make_ids(0, 6) + _make_ids(1, 8) + _make_ids(2, 8)

    def _read_port(self, port):
        return int(self._connection.execute("DIN(X,%d,%d)" % (port[0], port[1])))

    def _write_port(self, port, value):
        self._connection.execute("DOUT(X,%d,%d:%d)" % (port[0], port[1], value))



from concert.quantities import q
from concert.tests import assert_almost_equal, TestCase
from concert.devices.motors.dummy import LinearMotor
from concert.processes import scan, ascan, dscan, scan_param_feedback


def compare_sequences(first_sequence, second_sequence, assertion):
    for x, y in zip(first_sequence, second_sequence):
        assertion(x, y)


class TestScan(TestCase):

    def setUp(self):
        super(TestScan, self).setUp()
        self.motor = LinearMotor()

    def handle_scan(self, parameters):
        self.positions.append(parameters[0].get().result())

    def test_ascan(self):
        self.positions = []

        ascan([(self.motor['position'], -2 * q.mm, 10 * q.mm)],
              n_intervals=4,
              handler=self.handle_scan)

        expected = [-2 * q.mm, 1 * q.mm, 4 * q.mm, 7 * q.mm, 10 * q.mm]
        compare_sequences(self.positions, expected, assert_almost_equal)

    def test_dscan(self):
        self.positions = []

        self.motor.position = 2 * q.mm
        dscan([(self.motor['position'], 2 * q.mm, 10 * q.mm)],
              n_intervals=4,
              handler=self.handle_scan)

        expected = [4 * q.mm, 6 * q.mm, 8 * q.mm, 10 * q.mm, 12 * q.mm]
        compare_sequences(self.positions, expected, assert_almost_equal)

    def test_process(self):
        def feedback():
            return self.motor.position

        x, y = scan(self.motor['position'], feedback,
                    1 * q.mm, 10 * q.mm, 10).result()
        compare_sequences(x, y, self.assertEqual)

    def test_scan_param_feedback(self):
        p = self.motor['position']
        x, y = scan_param_feedback(p, p, 1 * q.mm, 10 * q.mm, 10).result()
        compare_sequences(x, y, self.assertEqual)



import numpy as np
from concert.coroutines.base import coroutine, broadcast, inject
from concert.coroutines.filters import (absorptivity, backproject, flat_correct, average_images,
                                        queue, sinograms, downsize, stall, PickSlice)
from concert.coroutines.sinks import null, Result
from concert.tests import TestCase


def generator():
    for i in range(5):
        yield i


def frame_producer(consumer, num_frames=3, shape=(2, 2)):
    for i in range(num_frames):
        consumer.send(np.ones(shape=shape) * (i + 1))


class TestCoroutines(TestCase):

    def setUp(self):
        super(TestCoroutines, self).setUp()
        self.data = None
        self.data_2 = None
        self.stack = []
        self.iteration = 0

    def producer(self, consumer, num_items=5):
        for i in range(num_items):
            consumer.send(i)
            self.iteration += 1

    @coroutine
    def consume(self):
        while True:
            self.data = yield

    @coroutine
    def consume_2(self):
        while True:
            self.data_2 = yield

    @coroutine
    def stack_consume(self):
        while True:
            item = yield
            self.stack.append(item)

    def test_broadcast(self):
        self.producer(broadcast(self.consume(), self.consume_2()))
        self.assertEqual(self.data, 4)
        self.assertEqual(self.data_2, 4)

    def test_sinogenerator(self):
        n = 4

        result = Result()
        frame_producer(sinograms(n, result()), num_frames=8, shape=(4, 4))

        sinos = result.result
        # One sinogram
        one = np.tile(np.array([5, 6, 7, 8])[:, np.newaxis], [1, n])
        np.testing.assert_almost_equal(sinos, np.tile(one, [n, 1, 1]))

    def test_injection(self):
        inject(generator(), self.consume())
        self.assertEqual(self.data, 4)

    def test_null(self):
        self.producer(null())
        self.assertEquals(5, self.iteration)

    def test_averager(self):
        frame_producer(average_images(self.consume()))
        truth = np.ones((2, 2)) * 2
        np.testing.assert_almost_equal(self.data, truth)

    def test_flat_correct(self):
        shape = (2, 2)
        dark = np.ones(shape)
        flat = np.ones(shape) * 10
        truth_base = np.ones(shape)

        @coroutine
        def check():
            i = 1
            while True:
                frame = yield
                value = (i - 1) / 9.0
                np.testing.assert_almost_equal(frame, truth_base * value)
                i += 1

        frame_producer(flat_correct(flat, check(), dark=dark))

    def test_absorptivity(self):
        truth_base = np.ones((2, 2))

        @coroutine
        def check():
            i = 1
            while True:
                frame = yield
                np.testing.assert_almost_equal(frame, -np.log(truth_base * i))
                i += 1
        frame_producer(absorptivity(check()))

    def test_result_object(self):
        result = Result()
        self.producer(result())
        self.assertEqual(result.result, 4)

    def test_downsize(self):
        frame_producer(downsize(self.stack_consume(), x_slice=(2, 6, 2), y_slice=(3, 10, 3),
                                z_slice=(2, 8, 3)), num_frames=10, shape=(10, 10))
        result = np.array(self.stack)
        ones = np.ones(shape=(3, 2))

        self.assertEqual(result.shape, (2, 3, 2))
        np.testing.assert_almost_equal(result[0], ones * 3)
        np.testing.assert_almost_equal(result[1], ones * 6)

    def test_stall(self):
        # Simple case, no modulo
        self.producer(stall(self.stack_consume(), per_shot=5), num_items=10)
        self.assertEqual(self.stack, [4, 9])

        # More iterations than total items
        self.stack = []
        self.producer(stall(self.stack_consume(), per_shot=8, flush_at=13), num_items=26)
        self.assertEqual(self.stack, [7, 12, 20, 25])

    def test_slicepick(self):
        def produce_volume(consumer):
            vol = np.ones((2, 2, 2))
            vol[1] *= 2
            consumer.send(vol)

        pick = PickSlice(0)

        produce_volume(pick(self.consume()))
        np.testing.assert_almost_equal(self.data, np.ones((2, 2)))

        pick.index = 1
        produce_volume(pick(self.consume()))
        np.testing.assert_almost_equal(self.data, np.ones((2, 2)) * 2)

    def test_queue(self):
        frame_producer(queue(null()))

    def test_backproject(self):
        frame_producer(backproject(1, null()))



import numpy as np
from nose.plugins.attrib import attr
from concert.quantities import q
from concert.devices.motors.dummy import ContinuousRotationMotor
from concert.processes import align_rotation_axis
from concert.tests import slow, TestCase
from concert.tests.util.rotationaxis import SimulationCamera


@attr('skip-travis')
class TestDummyAlignment(TestCase):

    def setUp(self):
        super(TestDummyAlignment, self).setUp()
        self.x_motor = ContinuousRotationMotor()
        self.y_motor = ContinuousRotationMotor()
        self.z_motor = ContinuousRotationMotor()

        self.x_motor.position = 0 * q.deg
        self.y_motor.position = 0 * q.deg
        self.z_motor.position = 0 * q.deg

        self.camera = SimulationCamera(128, self.x_motor["position"],
                                       self.y_motor["position"],
                                       self.z_motor["position"])

        # Allow 1 px misalignment in y-direction.
        self.eps = np.arctan(2.0 / self.camera.rotation_radius) * q.rad

    def align_check(self, x_angle, z_angle, has_x_motor=True,
                    has_z_motor=True, flat=None, dark=None):
        """"Align and check the results."""
        self.x_motor.position = z_angle
        self.z_motor.position = x_angle

        x_motor = self.x_motor if has_x_motor else None
        z_motor = self.z_motor if has_z_motor else None

        align_rotation_axis(self.camera, self.y_motor, x_motor=x_motor, z_motor=z_motor, flat=flat,
                dark=dark).join()

        # In our case the best perfectly aligned position is when both
        # motors are in 0.
        if has_x_motor:
            assert np.abs(self.x_motor.position) < self.eps
        if has_z_motor:
            assert np.abs(self.z_motor.position) < self.eps

    def test_no_motor(self):
        with self.assertRaises(ValueError):
            align_rotation_axis(self.camera, self.y_motor, x_motor=None, z_motor=None).join()

    @slow
    def test_out_of_fov(self):
        def get_ones():
            return np.ones((self.camera.size,
                            self.camera.size))

        self.camera._grab_real = get_ones

        with self.assertRaises(ValueError) as ctx:
            align_rotation_axis(self.camera, self.y_motor,
                                x_motor=self.x_motor,
                                z_motor=self.z_motor).join()

        self.assertEqual("No sample tip points found.", str(ctx.exception))

    @slow
    def test_not_offcentered(self):
        self.camera.rotation_radius = 0
        with self.assertRaises(ValueError) as ctx:
            align_rotation_axis(self.camera, self.y_motor,
                                x_motor=self.x_motor,
                                z_motor=self.z_motor).join()

        self.assertEqual("Sample off-centering too " +
                         "small, enlarge rotation radius.",
                         str(ctx.exception))

    @slow
    def test_no_x_axis(self):
        """Test the case when there is no x-axis motor available."""
        self.align_check(17 * q.deg, 11 * q.deg, has_z_motor=False)

    @slow
    def test_no_z_axis(self):
        """Test the case when there is no x-axis motor available."""
        self.align_check(17 * q.deg, 11 * q.deg, has_x_motor=False)

    @slow
    def test_not_misaligned(self):
        "Perfectly aligned rotation axis."
        self.align_check(0 * q.deg, 0 * q.deg)

    @slow
    def test_only_x(self):
        """Only misaligned laterally."""
        self.align_check(-17 * q.deg, 0 * q.deg)

    @slow
    def test_only_z(self):
        """Only misaligned in the beam direction."""
        self.align_check(0 * q.deg, 11 * q.deg)

    @slow
    def test_huge_x(self):
        self.camera.scale = (3, 0.25, 3)
        self.align_check(60 * q.deg, 11 * q.deg)

    @slow
    def test_huge_z(self):
        self.camera.scale = (3, 0.25, 3)
        self.align_check(11 * q.deg, 60 * q.deg)

    @slow
    def test_positive(self):
        self.align_check(17 * q.deg, 11 * q.deg)

    @slow
    def test_negative_positive(self):
        self.align_check(-17 * q.deg, 11 * q.deg)

    @slow
    def test_positive_negative(self):
        self.align_check(17 * q.deg, -11 * q.deg)

    @slow
    def test_negative(self):
        self.align_check(-17 * q.deg, -11 * q.deg)

    @slow
    def test_flat_corrected(self):
        shape = (self.camera.size, self.camera.size)
        dark = np.zeros(shape, dtype=np.float)
        flat = np.ones(shape, dtype=np.float)
        self.align_check(-17 * q.deg, -11 * q.deg, flat=flat, dark=dark)



import functools


def coroutine(func):
    """
    Start a coroutine automatically without the need to call
    next() or send(None) first.
    """
    @functools.wraps(func)
    def start(*args, **kwargs):
        """Starts the generator."""
        gen = func(*args, **kwargs)
        next(gen)
        return gen
    return start


def inject(generator, consumer):
    """
    Let a *generator* produce a value and forward it to *consumer*.
    """
    for item in generator:
        consumer.send(item)


@coroutine
def broadcast(*consumers):
    """
    broadcast(*consumers)

    Forward data to all *consumers*.
    """
    while True:
        item = yield
        for consumer in consumers:
            consumer.send(item)



"""New port 74000 monochromator"""

import time
from concert.quantities import q
from concert.networking import SocketConnection
from concert.devices.monochromators.base import Monochromator
from concert.devices.base import LinearCalibration


class NewPort74000(Monochromator):

    """NewPort 74000 mocochromator class implementation"""

    def __init__(self, host, port):

        self._connection = SocketConnection(host, port)
        calibration = LinearCalibration(q.count / q.nm, 0 * q.nm)
        super(NewPort74000, self).__init__(calibration)
        self.steps = 0

    def shutter_open(self):
        """Open shutter"""
        self._connection.send('SHUTTER O\r\n')

    def shutter_close(self):
        """Close shutter"""
        self._connection.send('SHUTTER C\r\n')

    def shutter_status(self):
        """Get the shutter status"""
        self._connection.send('SHUTTER?\r\n')
        time.sleep(0.1)
        status = self._connection.recv()[11]
        return status

    def _set_wavelength(self, wave):
        """Set the wavelength"""
        self._connection.send('GOWAVE ' + wave + '\r\n')

    def _get_wavelength(self):
        """Get the current wavelength"""
        self._connection.send('WAVE?\r\n')
        time.sleep(0.1)
        self.steps = self._connection.recv()[8:]
        return self.steps



"""Shutter Device."""
from concert.base import transition, State, AccessorNotImplementedError
from concert.async import async
from concert.devices.base import Device


class Shutter(Device):

    """Shutter device class implementation."""

    state = State(default="open")

    def __init__(self):
        super(Shutter, self).__init__()

    @async
    @transition(source='closed', target='open')
    def open(self):
        """open()

        Open the shutter."""
        self._open()

    @async
    @transition(source='open', target='closed')
    def close(self):
        """close()

        Close the shutter."""
        self._close()

    def _open(self):
        raise AccessorNotImplementedError

    def _close(self):
        raise AccessorNotImplementedError



from concert.quantities import q
from concert.tests import TestCase
from concert.base import (Parameterizable, Parameter, Quantity, State, SoftLimitError,
                          transition, LockError)


class BaseDevice(Parameterizable):

    def __init__(self):
        super(BaseDevice, self).__init__()


class FooDevice(BaseDevice):

    state = State(default='standby')

    foo = Quantity(q.m, transition=transition(source='*', target='moved'))

    def __init__(self, default):
        super(FooDevice, self).__init__()
        self._value = default

    def _get_foo(self):
        return self._value

    def _set_foo(self, value):
        self._value = value

    param = Parameter(fget=_get_foo, fset=_set_foo)


class RestrictedFooDevice(FooDevice):
    def __init__(self, lower, upper):
        super(RestrictedFooDevice, self).__init__(0 * q.mm)
        self['foo'].lower = lower
        self['foo'].upper = upper


class TestDescriptor(TestCase):

    def setUp(self):
        super(TestDescriptor, self).setUp()
        self.foo1 = FooDevice(42 * q.m)
        self.foo2 = FooDevice(23 * q.m)

    def test_property_identity(self):
        self.foo1.foo = 15 * q.m
        self.assertEqual(self.foo1.foo, 15 * q.m)
        self.assertEqual(self.foo2.foo, 23 * q.m)

    def test_func_identity(self):
        self.foo1.set_foo(15 * q.m).join()
        self.assertEqual(self.foo1.get_foo().result(), 15 * q.m)
        self.assertEqual(self.foo2.get_foo().result(), 23 * q.m)


class TestParameterizable(TestCase):

    def setUp(self):
        super(TestParameterizable, self).setUp()
        self.device = FooDevice(0 * q.mm)

    def test_param(self):
        self.device.param = 15
        self.assertEqual(self.device.param, 15)

    def test_saving(self):
        self.device.foo = 1 * q.mm
        self.device.stash().join()
        self.device.foo = 2 * q.mm
        self.device.stash().join()
        self.device.foo = 0.123 * q.mm
        self.device.foo = 1.234 * q.mm

        self.device.restore().join()
        self.assertEqual(self.device.foo, 2 * q.mm)

        self.device.restore().join()
        self.assertEqual(self.device.foo, 1 * q.mm)

    def test_state(self):
        self.assertEqual(self.device.state, 'standby')
        self.device.foo = 2 * q.m
        self.assertEqual(self.device.state, 'moved')

    def test_lock(self):
        self.device['foo'].lock()
        self.assertTrue(self.device['foo'].locked)

        with self.assertRaises(LockError):
            self.device.foo = 1 * q.m

        # This must pass
        self.device.param = 1 * q.m

        # Unlock and test if writable
        self.device['foo'].unlock()
        self.assertFalse(self.device['foo'].locked)
        self.device.foo = 1 * q.m

        # Lock the whole device
        self.device.lock()

        with self.assertRaises(LockError):
            self.device.param = 1 * q.m

        with self.assertRaises(LockError):
            self.device.foo = 1 * q.m

        # Unlock the whole device and test if writable
        self.device.unlock()
        self.device.param = 1 * q.m
        self.device.foo = 1 * q.m

    def test_permanent_lock(self):
        self.device['foo'].lock(permanent=True)
        self.assertTrue(self.device['foo'].locked)

        with self.assertRaises(LockError):
            self.device.foo = 1 * q.m

        with self.assertRaises(LockError):
            self.device['foo'].unlock()

    def test_permanent_device_lock(self):
        self.device.lock(permanent=True)

        with self.assertRaises(LockError):
            self.device.unlock()


class TestQuantity(TestCase):

    def test_soft_limit_change(self):
        limited = FooDevice(0 * q.mm)
        limited['foo'].lower = -2 * q.mm
        limited['foo'].upper = 2 * q.mm

        limited.foo = -1.5 * q.mm
        limited.foo = +1.5 * q.mm

        with self.assertRaises(SoftLimitError):
            limited.foo = 2.5 * q.mm

    def test_soft_limit_restriction(self):
        limited = RestrictedFooDevice(-2 * q.mm, 2 * q.mm)
        self.assertEqual(limited['foo'].lower, -2 * q.mm)
        self.assertEqual(limited['foo'].upper, +2 * q.mm)

    def test_parameter_property(self):
        device = FooDevice(42 * q.m)
        self.assertEqual(device['foo'].unit, q.m)

    def test_limits_lock(self):
        device = FooDevice(10 * q.mm)
        device['foo'].lock_limits()
        with self.assertRaises(LockError):
            device['foo'].lower = -10 * q.mm

        device['foo'].lock_limits(True)
        with self.assertRaises(LockError):
            device['foo'].unlock_limits()




"""EdmundOptics photodiode implementation."""

from concert.quantities import q
from concert.devices.io.wago import IO
from concert.devices.photodiodes import base


class PhotoDiode(base.PhotoDiode):

    """
    Impementation of Edmund Optics photodiode with V output signal
    """

    def __init__(self, host, port):
        self._host = host
        self._port = port
        super(PhotoDiode, self).__init__()

    def _get_intensity(self, port):
        """
        Read output intensity from the photodiode
        """
        connection = IO(self._host, self._port)
        intensity = float(connection._read_port(port)) / 1000
        return intensity * q.V



"""
Detector is a device which consists of a camera and an objective lens
which affects the effective pixel size.
"""
from concert.quantities import q
from concert.base import Quantity
from concert.devices.base import Device


class Detector(Device):
    """
    Detector consisting of a :class:`.concert.devices.cameras.base.Camera`
    and a magnification given by an objective lens.
    """

    pixel_width = Quantity(q.m, help="Effective pixel width")
    pixel_height = Quantity(q.m, help="Effective pixel height")

    def __init__(self, camera, magnification):
        super(Detector, self).__init__()
        self.camera = camera
        self.magnification = magnification

    def _get_pixel_width(self):
        return self.camera.sensor_pixel_width * self.magnification

    def _get_pixel_height(self):
        return self.camera.sensor_pixel_height * self.magnification



"""Opening images in external programs."""
import collections
import atexit
import os
import tempfile
import time
try:
    from Queue import Empty
except ImportError:
    from queue import Empty
import logging
import numpy as np
from subprocess import Popen
from multiprocessing import Queue as MultiprocessingQueue, Process
from concert.async import threaded
from concert.quantities import q
from concert.storage import write_tiff
from concert.coroutines.base import coroutine


LOG = logging.getLogger(__name__)
_PYPLOT_VIEWERS = []


def _terminate_pyplotviewers():
    """Terminate all :py:class:`PyplotViewerBase` isntances."""
    for viewer in _PYPLOT_VIEWERS:
        if viewer is not None:
            viewer.terminate()


# Register termination of all the PyplotViewerBase instances also on
# normal exit
atexit.register(_terminate_pyplotviewers)


def _start_command(program, image, writer=write_tiff):
    """
    Create a tmp file for dumping the *image* and use *program*
    to open that image. Use *writer* for writing the iamge to the disk.
    """
    tmp_file = tempfile.mkstemp()[1]
    try:
        full_path = writer(tmp_file, image)
        process = Popen([program, full_path])
        process.wait()
    finally:
        os.remove(full_path)


@threaded
def imagej(image, path="imagej", writer=write_tiff):
    """
    Open *image* in ImageJ found by *path*. *writer* specifies
    the written image file type.
    """
    _start_command(path, image, writer)


class PyplotViewerBase(object):

    """
    A base class for data viewer which sends commands to a matplotlib updater
    which runs in a separate process.

    .. py:attribute:: view_function

        The function which updates the figure based on the changed data. Its
        nomenclature has to be::

            foo(data, force=False)

        Where *force* determines whether the redrawing must be done or not. If
        it is False, the redrawing takes place if the data queue contains only
        the current data item. This prevents the actual drawer from being
        overwhelmed by the amount of incoming data.

    .. py:attribute:: blit

        True if faster redrawing based on canvas blitting should be used.
    """

    def __init__(self, view_function, blit=False):
        self._queue = MultiprocessingQueue()
        self._paused = False
        self._terminated = False
        self._coroutine = None
        self._proc = None
        # The udater is implementation-specific and must be provided by
        # the subclass by calling self._set_updater
        self._updater = None
        self.view_function = view_function
        self._blit = blit

    def __call__(self, size=None):
        """
        Display a dynamic image in a separate thread in order not to
        stall program execution. If *size* is specified, the redrawing stops
        when *size* images come.
        """
        if self._coroutine is None:
            self._coroutine = self._update(size=size)

        return self._coroutine

    @coroutine
    def _update(self, size=None):
        """
        Display data image in a separate thread in order not to stall program
        execution. If *size* is specified, the redrawing stops when *size*
        data items come. This method does not force the redrawing unless the
        last item has come. The subclass must provide the view function
        :py:attr:`PyplotViewerBase.view_function`.
        """
        i = 0

        while True:
            item = yield
            if size is None or i < size:
                if size is not None and i == size - 1:
                    # Maximum number of items has come, end redrawing
                    self.view_function(item, force=True)
                else:
                    self.view_function(item)
                # This helps with the smoothness of drawing
                time.sleep(0.001)

            i += 1

    def _set_updater(self, updater):
        """
        Set the *updater*, now the process can start. This has to be called
        by the subclasses.
        """
        self._updater = updater
        self._proc = Process(target=self._run)
        self._proc.start()
        _PYPLOT_VIEWERS.append(self)

    def terminate(self):
        """Close all communication and terminate child process."""
        if not self._terminated:
            self._queue.close()
            self._proc.terminate()

    def pause(self):
        """Pause, no images are dispayed but image commands work."""
        self._paused = True

    def resume(self):
        """Resume the viewer."""
        self._paused = False

    def _run(self):
        """
        Run the process, i.e. wait for an image to come to the queue
        and dispaly it. This method is executed in a separate process.
        """
        # This import *must* be here, otherwise it doesn't work on Linux
        from matplotlib import pyplot as plt
        from matplotlib.animation import FuncAnimation

        try:
            figure = plt.figure()
            # The underscore must stay, otherwise it doesn't work, magic?
            _ = FuncAnimation(figure, self._updater.process, interval=5,
                              blit=self._blit)
            plt.show()

        except KeyboardInterrupt:
            plt.close("all")


class PyplotViewer(PyplotViewerBase):

    """
    Dynamic plot viewer using matplotlib.

    .. py:attribute:: style

        One of matplotlib's linestyle format strings

    .. py:attribute:: plt_kwargs

        Keyword arguments accepted by matplotlib's plot()

    .. py:attribute:: autoscale

        If True, the axes limits will be expanded as needed by the new data,
        otherwise the user needs to rescale the axes

    """

    def __init__(self, style="o", plot_kwargs=None, autoscale=True, title=""):
        super(PyplotViewer, self).__init__(self.plot)
        self._autoscale = autoscale
        self._style = style
        self._iteration = 0
        self._set_updater(_PyplotUpdater(self._queue, style,
                                         plot_kwargs, autoscale,
                                         title=title))

    def plot(self, x, y=None, force=False):
        """
        Plot *x* and *y*, if *y* is None and *x* is a scalar the real y is
        given by *x* and x is the current iteration of the plotting command,
        if *x* is an iterable then it is interpreted as y data array and x is
        a span [0, len(x)]. If both *x* and *y* are given, they are plotted as
        they are. If *force* is True the plotting is guaranteed, otherwise it
        might be skipped for the sake of plotting speed.

        Note: if x is not given, the iteration starts at 0.
        """
        if not self._paused and (self._queue.empty() or force):
            if y is None:
                if isinstance(x, q.Quantity) and isinstance(x.magnitude, collections.Iterable) or\
                   not isinstance(x, q.Quantity) and isinstance(x, collections.Iterable):
                    x_data = np.arange(len(x))
                    y_data = x
                else:
                    x_data = self._iteration
                    y_data = x
                    self._iteration += 1
            else:
                x_data = x
                y_data = y
            self._queue.put((_PyplotUpdater.PLOT, (x_data, y_data)))

    @property
    def style(self):
        return self._style

    @style.setter
    def style(self, style):
        """Set line style to *style*."""
        self._style = style
        self._queue.put((_PyplotUpdater.STYLE, style))

    def clear(self):
        """Clear the plotted data."""
        self._iteration = 0
        self._queue.put((_PyplotUpdater.CLEAR, None))

    @property
    def autoscale(self):
        return self._autoscale

    @autoscale.setter
    def autoscale(self, autoscale):
        """Set *autoscale* on the axes, can be True or False."""
        self._autoscale = autoscale
        self._queue.put((_PyplotUpdater.AUTOSCALE, autoscale))


class PyplotImageViewer(PyplotViewerBase):

    """Dynamic image viewer using matplotlib."""

    def __init__(self, imshow_kwargs=None, colorbar=True, title=""):
        super(PyplotImageViewer, self).__init__(self.show, blit=True)
        self._has_colorbar = colorbar
        self._imshow_kwargs = {} if imshow_kwargs is None else imshow_kwargs
        self._make_imshow_defaults()
        self._set_updater(_PyplotImageUpdater(self._queue,
                                              self._imshow_kwargs,
                                              self._has_colorbar,
                                              title=title))

    def show(self, item, force=False):
        """
        show *item* into the redrawing queue. The item is truly inserted
        only if the queue is empty in order to guarantee that the newest
        image is drawn or if the *force* is True.
        """
        if not self._paused and (self._queue.empty() or force):
            self._queue.put((_PyplotImageUpdater.IMAGE, item))

    @property
    def limits(self):
        raise NotImplementedError

    @limits.setter
    def limits(self, clim):
        """
        Update the colormap limits by *clim*, which is a (lower, upper)
        tuple.
        """
        self._queue.put((_PyplotImageUpdater.CLIM, clim))

    @property
    def colormap(self):
        return self._colormap

    @colormap.setter
    def colormap(self, colormap):
        """Set colormp of the shown image to *colormap*."""
        self._queue.put((_PyplotImageUpdater.COLORMAP, colormap))

    def _make_imshow_defaults(self):
        """Override matplotlib's image showing defafults."""
        from matplotlib import cm

        if "cmap" not in self._imshow_kwargs:
            self._imshow_kwargs["cmap"] = cm.gray
            self._colormap = cm.gray
        else:
            self._colormap = self._imshow_kwargs["cmap"]
        if "interpolation" not in self._imshow_kwargs:
            self._imshow_kwargs["interpolation"] = "nearest"


class _PyplotUpdaterBase(object):

    """
    Base class for animating a matploblib figure in a separate process.

    .. py:attribute:: queue

        A multiprocessing queue for receiving commands
    """

    def __init__(self, queue, title=""):
        self.queue = queue
        self.first = True
        self.title = title
        # A dictionary in form command: method which tells the class what to do
        # for every received command
        self.commands = {}

    def process(self, iteration):
        """
        Update function which is going to be called by matplotlib's
        FuncAnimation instance.
        """
        try:
            if self.first:
                # Wait as much time as it takes for the first
                # time beacuse we don't want to show a window
                # with no image in it.
                item = self.queue.get()
                self.first = False
            else:
                item = self.queue.get(timeout=0.1)
            command, data = item
            self.commands[command](data)
        except Empty:
            pass
        except Exception as exc:
            LOG.error("Pyplot Exception: {}".format(exc))
        finally:
            return self.get_artists()

    def get_artists(self):
        """
        Abstract function for getting all matplotlib artists which we want
        to redraw. Needs to be implemented by the subclass.
        """
        raise NotImplementedError


class _PyplotUpdater(_PyplotUpdaterBase):

    """
    Private class for updating a matplotlib figure with a 1D data stream.
    The arguments are the same as by :py:class:`PyplotViewer`.
    """
    CLEAR = "clear"
    PLOT = "plot"
    STYLE = "style"
    AUTOSCALE = "autoscale"

    def __init__(self, queue, style="o", plot_kwargs=None, autoscale=True,
                 title=""):
        super(_PyplotUpdater, self).__init__(queue, title=title)
        self.data = [[], []]
        self.line = None
        self.style = style
        self.plot_kwargs = {} if plot_kwargs is None else plot_kwargs
        self.autoscale = autoscale
        self.commands = {_PyplotUpdater.PLOT: self.plot,
                         _PyplotUpdater.STYLE: self.change_style,
                         _PyplotUpdater.CLEAR: self.clear,
                         _PyplotUpdater.AUTOSCALE: self.set_autoscale}

    def get_artists(self):
        """Get the artists for the drawing."""
        return [] if self.line is None else [self.line]

    def make_line(self):
        """Create the line based on current settings."""
        from matplotlib import pyplot as plt

        self.line = plt.plot(self.data[0], self.data[1], self.style,
                             **self.plot_kwargs)[0]
        self.line.axes.set_title(self.title)

    def clear(self, data):
        """Clear everything from the plot."""
        from matplotlib import pyplot as plt

        self.data = [[], []]
        if self.line is not None:
            self.line.axes.clear()
        self.make_line()

        plt.draw()

    def plot(self, data):
        """Plot *data*, which is an (x, y) tuple."""
        from matplotlib import pyplot as plt

        def get_magnitude_and_unit(value):
            if hasattr(value, "magnitude"):
                dimless = value.magnitude
                unit = value.units
            else:
                dimless = value
                unit = None

            return dimless, unit

        x_item, x_units = get_magnitude_and_unit(data[0])
        y_item, y_units = get_magnitude_and_unit(data[1])

        first = len(self.data[0]) == 0
        is_iterable = isinstance(x_item, collections.Iterable)
        if is_iterable:
            self.data = [list(x_item), list(y_item)]
        else:
            self.data[0].append(x_item)
            self.data[1].append(y_item)

        if first:
            self.make_line()
            if not is_iterable:
                # Make sure the limits are set properly for the first time
                self.line.axes.set_xlim(self.data[0][0] - 1e-7, self.data[0][0] + 1e-7)
                self.line.axes.set_ylim(self.data[1][0] - 1e-7, self.data[1][0] + 1e-7)
            if x_units is not None:
                self.line.axes.get_xaxis().set_label_text(str(x_units))
            if y_units is not None:
                self.line.axes.get_yaxis().set_label_text(str(y_units))

        self.line.set_data(*self.data)

        if self.autoscale or first:
            # Draw for the first time or on limit change to update ticks and
            # labels
            self.autoscale_view()
        plt.draw()

    def change_style(self, style):
        """Change line style to *style*."""
        self.style = style
        if self.line is not None:
            # Just redrawing doesn't work
            self.line.axes.clear()
            self.make_line()

    def set_autoscale(self, autoscale):
        """If *autoscale* is True, the plit is rescaled when needed."""
        from matplotlib import pyplot as plt

        self.autoscale = autoscale
        if self.autoscale:
            self.autoscale_view()
            plt.draw()

    def autoscale_view(self):
        """Autoscale axes limits."""
        if self.line is not None:
            self.line.axes.relim()
            # For some reason the relim itself doesn't work, so we set the
            # limits to the new values explicitly
            if len(self.data[0]) > 1:
                self.line.axes.set_xlim(min(self.data[0]), max(self.data[0]))
                self.line.axes.set_ylim(min(self.data[1]), max(self.data[1]))
            self.line.axes.autoscale_view()


class _PyplotImageUpdater(_PyplotUpdaterBase):

    """
    Private class for updating a matplotlib figure with an image stream.
    """

    IMAGE = "image"
    CLIM = "clim"
    COLORMAP = "colormap"

    def __init__(self, queue, imshow_kwargs, has_colorbar, title=""):
        super(_PyplotImageUpdater, self).__init__(queue, title=title)
        self.imshow_kwargs = imshow_kwargs
        self.has_colorbar = has_colorbar
        self.mpl_image = None
        self.colorbar = None
        self.clim = None
        self.colormap = None
        self.commands = {_PyplotImageUpdater.IMAGE: self.process_image,
                         _PyplotImageUpdater.CLIM: self.update_limits,
                         _PyplotImageUpdater.COLORMAP: self.update_colormap}

    def get_artists(self):
        """Get artists to return for matplotlib's animation."""
        retval = [] if self.mpl_image is None else [self.mpl_image]

        return retval

    def process_image(self, image):
        """Process the incoming *image*."""
        if self.mpl_image is not None and \
                self.mpl_image.get_size() != image.shape:
            # When the shape changes the axes needs to be reset
            self.mpl_image.axes.clear()
            self.mpl_image = None

        if self.mpl_image is None:
            # Either removed by shape change or first time drawing
            self.make_image(image)
        else:
            self.update_all(image)

    def update_limits(self, clim):
        """
        Update current colormap limits by *clim*, which is a tuple
        (lower, upper). If *clim* is None, the limit is reset to the span of
        the current image.
        """
        from matplotlib import pyplot as plt
        self.clim = clim

        if clim is None and self.mpl_image is not None:
            # Restore the full clim
            clim = self.mpl_image.get_array().min(), \
                self.mpl_image.get_array().max()

        if self.mpl_image is not None and clim is not None:
            self.mpl_image.set_clim(clim)
            self.update_colorbar()
            plt.draw()

    def update_colormap(self, colormap):
        """Update colormap to *colormap*."""
        from matplotlib import pyplot as plt

        self.colormap = colormap

        if self.mpl_image is not None:
            self.mpl_image.set_cmap(self.colormap)
        if "cmap" in self.imshow_kwargs:
            self.imshow_kwargs["cmap"] = self.colormap
        if self.colorbar is not None:
            self.colorbar.set_cmap(self.colormap)
            self.colorbar.draw_all()
        plt.draw()

    def update_all(self, image):
        """Update image and colorbar."""
        from matplotlib import pyplot as plt

        self.mpl_image.set_data(image)
        if self.clim is None:
            # If the limit is not set to a value we autoscale
            new_lower = float(image.min())
            new_upper = float(image.max())
            if self.limits_changed(new_lower, new_upper):
                self.mpl_image.set_clim(new_lower, new_upper)
                self.update_colorbar()
                plt.draw()

    def update_colorbar(self):
        """Update the colorbar (rescale and redraw)."""
        self.colorbar.set_clim(self.mpl_image.get_clim())
        self.colorbar.draw_all()

    def make_colorbar(self):
        """Make colorbar according to the current colormap."""
        from matplotlib import pyplot as plt

        colormap = None
        if self.colormap is not None:
            colormap = self.colormap
        elif "cmap" in self.imshow_kwargs:
            colormap = self.imshow_kwargs["cmap"]

        self.colorbar = plt.colorbar(cmap=colormap)

    def make_image(self, image):
        """Create an image with colorbar"""
        from matplotlib import pyplot as plt
        self.mpl_image = plt.imshow(image, **self.imshow_kwargs)
        self.mpl_image.axes.set_title(self.title)
        if self.clim is not None:
            self.mpl_image.set_clim(self.clim)
        if self.has_colorbar:
            if self.colorbar is None:
                self.make_colorbar()
            else:
                self.update_colorbar()
        plt.draw()

    def limits_changed(self, lower, upper):
        """
        Determine whether the colormap limits changed enough for colorbar
        redrawing.
        """
        new_range = upper - lower
        lower_ratio = np.abs(lower - self.mpl_image.get_clim()[0]) / new_range
        upper_ratio = np.abs(upper - self.mpl_image.get_clim()[1]) / new_range

        return lower_ratio > 0.1 or upper_ratio > 0.1



"""This module provides a simple dummy camera."""
import os
import time
import numpy as np
from concert.quantities import q
from concert.base import Quantity
from concert.devices.cameras import base
from concert.storage import read_image


class Base(base.Camera):

    exposure_time = Quantity(q.s)
    sensor_pixel_width = Quantity(q.micrometer)
    sensor_pixel_height = Quantity(q.micrometer)
    roi_x0 = Quantity(q.pixel)
    roi_y0 = Quantity(q.pixel)
    roi_width = Quantity(q.pixel)
    roi_height = Quantity(q.pixel)

    def __init__(self):
        super(Base, self).__init__()
        self._frame_rate = 1000 / q.s
        self._trigger_mode = self.trigger_modes.AUTO
        self._exposure_time = 1 * q.ms
        self._roi_x0 = 0 * q.pixel
        self._roi_y0 = 0 * q.pixel
        self._roi_width = 640 * q.pixel
        self._roi_height = 480 * q.pixel

    def _get_sensor_pixel_width(self):
        return 5 * q.micrometer

    def _get_sensor_pixel_height(self):
        return 5 * q.micrometer

    def _get_roi_x0(self):
        return self._roi_x0

    def _set_roi_x0(self, x0):
        self._roi_x0 = x0

    def _get_roi_y0(self):
        return self._roi_y0

    def _set_roi_y0(self, y0):
        self._roi_y0 = y0

    def _get_roi_width(self):
        return self._roi_width

    def _set_roi_width(self, roi):
        self._roi_width = roi

    def _get_roi_height(self):
        return self._roi_height

    def _set_roi_height(self, roi):
        self._roi_height = roi

    def _get_exposure_time(self):
        return self._exposure_time

    def _set_exposure_time(self, value):
        self._exposure_time = value

    def _get_frame_rate(self):
        return self._frame_rate

    def _set_frame_rate(self, frame_rate):
        self._frame_rate = frame_rate

    def _get_trigger_mode(self):
        return self._trigger_mode

    def _set_trigger_mode(self, mode):
        self._trigger_mode = mode

    def _record_real(self):
        pass

    def _stop_real(self):
        pass

    def _trigger_real(self):
        pass


class Camera(Base):

    """A simple dummy camera."""

    def __init__(self, background=None):
        """
        *background* can be an array-like that will be used to generate the frame
        when calling :meth:`.grab`. The final image will be the background +
        poisson noise depending on the currently set exposure time.
        """
        super(Camera, self).__init__()

        if background is not None:
            self.roi_width = background.shape[1] * q.pixel
            self.roi_height = background.shape[0] * q.pixel
            self._background = background
        else:
            shape = (640, 480)
            self.roi_width, self.roi_height = shape * q.pixel
            self._background = np.ones(shape)

    def _grab_real(self):
        start = time.time()
        cur_time = self.exposure_time.to(q.s).magnitude
        # 1e5 is a dummy correlation between exposure time and emitted e-.
        tmp = self._background + cur_time * 1e5
        max_value = np.iinfo(np.uint16).max
        tmp = np.random.poisson(tmp)
        # Cut values beyond the bit-depth.
        tmp[tmp > max_value] = max_value
        duration = time.time() - start
        to_sleep = 1.0 / self.frame_rate
        to_sleep = to_sleep.to_base_units() - duration * q.s
        if to_sleep > 0 * q.s:
            time.sleep(to_sleep.magnitude)

        return np.cast[np.uint16](tmp)


class FileCamera(Base):

    """A camera that reads files in a *directory*."""

    def __init__(self, directory):
        # Let users change the directory
        self.directory = directory
        super(FileCamera, self).__init__()

        self._index = 0
        self._files = [os.path.join(directory, file_name) for file_name in
                       sorted(os.listdir(directory))]

        if not self._files:
            raise base.CameraError("No files found")

        image = read_image(self._files[0])
        self._roi_width = image.shape[1] * q.pixel
        self._roi_height = image.shape[0] * q.pixel

    def _record_real(self):
        self._index = 0

    def _grab_real(self):
        if self._index < len(self._files):
            image = read_image(self._files[self._index])

            y_region = self.roi_y0 + self.roi_height
            x_region = self.roi_x0 + self.roi_width

            result = image[self.roi_y0.magnitude:y_region.magnitude,
                           self.roi_x0.magnitude:x_region.magnitude]
            self._index += 1
        else:
            result = None

        return result


class BufferedCamera(Camera, base.BufferedMixin):

    def _readout_real(self):
        for i in range(3):
            yield self.grab()
